
##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid
from pylipd.classes.archivetype import ArchiveType
from pylipd.classes.changelog import ChangeLog
from pylipd.classes.chrondata import ChronData
from pylipd.classes.funding import Funding
from pylipd.classes.location import Location
from pylipd.classes.paleodata import PaleoData
from pylipd.classes.person import Person
from pylipd.classes.publication import Publication

class Dataset:
    """Auto-generated LinkedEarth class representing `Dataset`."""
    def __init__(self):
        """Initialize a new Dataset instance."""
        
        self.archiveType: ArchiveType = None
        self.changeLogs: list[ChangeLog] = []
        self.chronData: list[ChronData] = []
        self.collectionName: str = None
        self.collectionYear: str = None
        self.compilationNest: str = None
        self.contributor: Person = None
        self.creators: list[Person] = []
        self.dataSource: str = None
        self.datasetId: str = None
        self.fundings: list[Funding] = []
        self.investigators: list[Person] = []
        self.location: Location = None
        self.name: str = None
        self.notes: str = None
        self.originalDataUrl: str = None
        self.paleoData: list[PaleoData] = []
        self.publications: list[Publication] = []
        self.spreadsheetLink: str = None
        self.version: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Dataset"
        self.id = self.ns + "/" + uniqid("Dataset.")

    @staticmethod
    def from_data(id, data) -> 'Dataset':
        """Instantiate `Dataset` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        Dataset
            The populated `Dataset` instance.
        """
        self = Dataset()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasArchiveType":
                for val in value:
                    obj = ArchiveType.from_synonym(re.sub("^.*?#", "", val["@id"]))
                                    
                    self.archiveType = obj
        
            elif key == "hasChangeLog":
                for val in value:
                    if "@id" in val:
                        obj = ChangeLog.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.changeLogs.append(obj)
        
            elif key == "hasChronData":
                for val in value:
                    if "@id" in val:
                        obj = ChronData.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.chronData.append(obj)
        
            elif key == "hasCollectionName":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.collectionName = obj
        
            elif key == "hasCollectionYear":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.collectionYear = obj
        
            elif key == "hasCompilationNest":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.compilationNest = obj
        
            elif key == "hasContributor":
                for val in value:
                    if "@id" in val:
                        obj = Person.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
                                    
                    self.contributor = obj
        
            elif key == "hasCreator":
                for val in value:
                    if "@id" in val:
                        obj = Person.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.creators.append(obj)
        
            elif key == "hasDataSource":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.dataSource = obj
        
            elif key == "hasDatasetId":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.datasetId = obj
        
            elif key == "hasFunding":
                for val in value:
                    if "@id" in val:
                        obj = Funding.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.fundings.append(obj)
        
            elif key == "hasInvestigator":
                for val in value:
                    if "@id" in val:
                        obj = Person.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.investigators.append(obj)
        
            elif key == "hasLocation":
                for val in value:
                    if "@id" in val:
                        obj = Location.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
                                    
                    self.location = obj
        
            elif key == "hasName":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.name = obj
        
            elif key == "hasNotes":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.notes = obj
        
            elif key == "hasOriginalDataUrl":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.originalDataUrl = obj
        
            elif key == "hasPaleoData":
                for val in value:
                    if "@id" in val:
                        obj = PaleoData.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.paleoData.append(obj)
        
            elif key == "hasPublication":
                for val in value:
                    if "@id" in val:
                        obj = Publication.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.publications.append(obj)
        
            elif key == "hasSpreadsheetLink":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.spreadsheetLink = obj
        
            elif key == "hasVersion":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.version = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if len(self.changeLogs):
            data[self.id]["hasChangeLog"] = []
        for value_obj in self.changeLogs:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasChangeLog"].append(obj)

        if len(self.chronData):
            data[self.id]["hasChronData"] = []
        for value_obj in self.chronData:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasChronData"].append(obj)

        if len(self.creators):
            data[self.id]["hasCreator"] = []
        for value_obj in self.creators:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasCreator"].append(obj)

        if len(self.fundings):
            data[self.id]["hasFunding"] = []
        for value_obj in self.fundings:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasFunding"].append(obj)

        if len(self.investigators):
            data[self.id]["hasInvestigator"] = []
        for value_obj in self.investigators:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasInvestigator"].append(obj)

        if len(self.paleoData):
            data[self.id]["hasPaleoData"] = []
        for value_obj in self.paleoData:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasPaleoData"].append(obj)

        if len(self.publications):
            data[self.id]["hasPublication"] = []
        for value_obj in self.publications:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasPublication"].append(obj)

        if self.archiveType:
            value_obj = self.archiveType
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasArchiveType"] = [obj]
                

        if self.collectionName:
            value_obj = self.collectionName
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCollectionName"] = [obj]
                

        if self.collectionYear:
            value_obj = self.collectionYear
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCollectionYear"] = [obj]
                

        if self.compilationNest:
            value_obj = self.compilationNest
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCompilationNest"] = [obj]
                

        if self.contributor:
            value_obj = self.contributor
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasContributor"] = [obj]
                

        if self.dataSource:
            value_obj = self.dataSource
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDataSource"] = [obj]
                

        if self.datasetId:
            value_obj = self.datasetId
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDatasetId"] = [obj]
                

        if self.location:
            value_obj = self.location
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasLocation"] = [obj]
                

        if self.name:
            value_obj = self.name
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasName"] = [obj]
                

        if self.notes:
            value_obj = self.notes
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasNotes"] = [obj]
                

        if self.originalDataUrl:
            value_obj = self.originalDataUrl
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasOriginalDataUrl"] = [obj]
                

        if self.spreadsheetLink:
            value_obj = self.spreadsheetLink
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasSpreadsheetLink"] = [obj]
                

        if self.version:
            value_obj = self.version
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasVersion"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if len(self.changeLogs):
            data["changelog"] = []
        for value_obj in self.changeLogs:
            obj = value_obj.to_json()
            data["changelog"].append(obj)

        if len(self.chronData):
            data["chronData"] = []
        for value_obj in self.chronData:
            obj = value_obj.to_json()
            data["chronData"].append(obj)

        if len(self.creators):
            data["creator"] = []
        for value_obj in self.creators:
            obj = value_obj.to_json()
            data["creator"].append(obj)

        if len(self.fundings):
            data["funding"] = []
        for value_obj in self.fundings:
            obj = value_obj.to_json()
            data["funding"].append(obj)

        if len(self.investigators):
            data["investigator"] = []
        for value_obj in self.investigators:
            obj = value_obj.to_json()
            data["investigator"].append(obj)

        if len(self.paleoData):
            data["paleoData"] = []
        for value_obj in self.paleoData:
            obj = value_obj.to_json()
            data["paleoData"].append(obj)

        if len(self.publications):
            data["pub"] = []
        for value_obj in self.publications:
            obj = value_obj.to_json()
            data["pub"].append(obj)

        if self.archiveType:
            value_obj = self.archiveType
            obj = value_obj.to_json()
            data["archiveType"] = obj

        if self.collectionName:
            value_obj = self.collectionName
            obj = value_obj
            data["collectionName"] = obj

        if self.collectionYear:
            value_obj = self.collectionYear
            obj = value_obj
            data["collectionYear"] = obj

        if self.compilationNest:
            value_obj = self.compilationNest
            obj = value_obj
            data["compilation_nest"] = obj

        if self.contributor:
            value_obj = self.contributor
            obj = value_obj.to_json()
            data["dataContributor"] = obj

        if self.dataSource:
            value_obj = self.dataSource
            obj = value_obj
            data["dataSource"] = obj

        if self.datasetId:
            value_obj = self.datasetId
            obj = value_obj
            data["datasetId"] = obj

        if self.location:
            value_obj = self.location
            obj = value_obj.to_json()
            data["geo"] = obj

        if self.name:
            value_obj = self.name
            obj = value_obj
            data["dataSetName"] = obj

        if self.notes:
            value_obj = self.notes
            obj = value_obj
            data["notes"] = obj

        if self.originalDataUrl:
            value_obj = self.originalDataUrl
            obj = value_obj
            data["originalDataURL"] = obj

        if self.spreadsheetLink:
            value_obj = self.spreadsheetLink
            obj = value_obj
            data["googleSpreadSheetKey"] = obj

        if self.version:
            value_obj = self.version
            obj = value_obj
            data["dataSetVersion"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Dataset':
        """Instantiate `Dataset` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        Dataset
            The populated `Dataset` instance.
        """
        self = Dataset()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "archiveType":
                    value = pvalue
                    obj = ArchiveType.from_synonym(re.sub("^.*?#", "", value))
                    self.archiveType = obj
            elif key == "changelog":
                for value in pvalue:
                    obj = ChangeLog.from_json(value)
                    self.changeLogs.append(obj)
            elif key == "chronData":
                for value in pvalue:
                    obj = ChronData.from_json(value)
                    self.chronData.append(obj)
            elif key == "collectionName":
                    value = pvalue
                    obj = value
                    self.collectionName = obj
            elif key == "collectionYear":
                    value = pvalue
                    obj = value
                    self.collectionYear = obj
            elif key == "compilation_nest":
                    value = pvalue
                    obj = value
                    self.compilationNest = obj
            elif key == "creator":
                for value in pvalue:
                    obj = Person.from_json(value)
                    self.creators.append(obj)
            elif key == "dataContributor":
                    value = pvalue
                    obj = Person.from_json(value)
                    self.contributor = obj
            elif key == "dataSetName":
                    value = pvalue
                    obj = value
                    self.name = obj
            elif key == "dataSetVersion":
                    value = pvalue
                    obj = value
                    self.version = obj
            elif key == "dataSource":
                    value = pvalue
                    obj = value
                    self.dataSource = obj
            elif key == "datasetId":
                    value = pvalue
                    obj = value
                    self.datasetId = obj
            elif key == "funding":
                for value in pvalue:
                    obj = Funding.from_json(value)
                    self.fundings.append(obj)
            elif key == "geo":
                    value = pvalue
                    obj = Location.from_json(value)
                    self.location = obj
            elif key == "googleSpreadSheetKey":
                    value = pvalue
                    obj = value
                    self.spreadsheetLink = obj
            elif key == "investigator":
                for value in pvalue:
                    obj = Person.from_json(value)
                    self.investigators.append(obj)
            elif key == "notes":
                    value = pvalue
                    obj = value
                    self.notes = obj
            elif key == "originalDataURL":
                    value = pvalue
                    obj = value
                    self.originalDataUrl = obj
            elif key == "paleoData":
                for value in pvalue:
                    obj = PaleoData.from_json(value)
                    self.paleoData.append(obj)
            elif key == "pub":
                for value in pvalue:
                    obj = Publication.from_json(value)
                    self.publications.append(obj)
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getArchiveType(self) -> ArchiveType:
        """Get archiveType.

        Returns
        -------
        ArchiveType
            The current value of archiveType.
        """
        return self.archiveType

    def setArchiveType(self, archiveType:ArchiveType):
        """Set archiveType.

        Parameters
        ----------
        archiveType : ArchiveType
            The value to assign.
        """
        assert isinstance(archiveType, ArchiveType), f"Error: '{archiveType}' is not of type ArchiveType\nYou can create a new ArchiveType object from a string using the following syntax:\n- Fetch existing ArchiveType by synonym: ArchiveType.from_synonym(\"{archiveType}\")\n- Create a new custom ArchiveType: ArchiveType(\"{archiveType}\")"
        self.archiveType = archiveType
    
    def getChangeLogs(self) -> list[ChangeLog]:
        """Get changeLogs list.

        Returns
        -------
        list[ChangeLog]
            A list of ChangeLog objects.
        """
        return self.changeLogs

    def setChangeLogs(self, changeLogs:list[ChangeLog]):
        """Set the changeLogs list.

        Parameters
        ----------
        changeLogs : list[ChangeLog]
            The list to assign.
        """
        assert isinstance(changeLogs, list), "Error: changeLogs is not a list"
        assert all(isinstance(x, ChangeLog) for x in changeLogs), f"Error: '{changeLogs}' is not of type ChangeLog"
        self.changeLogs = changeLogs

    def addChangeLog(self, changeLogs:ChangeLog):
        """Add a value to the changeLogs list.

        Parameters
        ----------
        changeLogs : ChangeLog
            The value to append.
        """
        assert isinstance(changeLogs, ChangeLog), f"Error: '{changeLogs}' is not of type ChangeLog"
        self.changeLogs.append(changeLogs)
        
    def getChronData(self) -> list[ChronData]:
        """Get chronData list.

        Returns
        -------
        list[ChronData]
            A list of ChronData objects.
        """
        return self.chronData

    def setChronData(self, chronData:list[ChronData]):
        """Set the chronData list.

        Parameters
        ----------
        chronData : list[ChronData]
            The list to assign.
        """
        assert isinstance(chronData, list), "Error: chronData is not a list"
        assert all(isinstance(x, ChronData) for x in chronData), f"Error: '{chronData}' is not of type ChronData"
        self.chronData = chronData

    def addChronData(self, chronData:ChronData):
        """Add a value to the chronData list.

        Parameters
        ----------
        chronData : ChronData
            The value to append.
        """
        assert isinstance(chronData, ChronData), f"Error: '{chronData}' is not of type ChronData"
        self.chronData.append(chronData)
        
    def getCollectionName(self) -> str:
        """Get collectionName.

        Returns
        -------
        str
            The current value of collectionName.
        """
        return self.collectionName

    def setCollectionName(self, collectionName:str):
        """Set collectionName.

        Parameters
        ----------
        collectionName : str
            The value to assign.
        """
        assert isinstance(collectionName, str), f"Error: '{collectionName}' is not of type str"
        self.collectionName = collectionName
    
    def getCollectionYear(self) -> str:
        """Get collectionYear.

        Returns
        -------
        str
            The current value of collectionYear.
        """
        return self.collectionYear

    def setCollectionYear(self, collectionYear:str):
        """Set collectionYear.

        Parameters
        ----------
        collectionYear : str
            The value to assign.
        """
        assert isinstance(collectionYear, str), f"Error: '{collectionYear}' is not of type str"
        self.collectionYear = collectionYear
    
    def getCompilationNest(self) -> str:
        """Get compilationNest.

        Returns
        -------
        str
            The current value of compilationNest.
        """
        return self.compilationNest

    def setCompilationNest(self, compilationNest:str):
        """Set compilationNest.

        Parameters
        ----------
        compilationNest : str
            The value to assign.
        """
        assert isinstance(compilationNest, str), f"Error: '{compilationNest}' is not of type str"
        self.compilationNest = compilationNest
    
    def getContributor(self) -> Person:
        """Get contributor.

        Returns
        -------
        Person
            The current value of contributor.
        """
        return self.contributor

    def setContributor(self, contributor:Person):
        """Set contributor.

        Parameters
        ----------
        contributor : Person
            The value to assign.
        """
        assert isinstance(contributor, Person), f"Error: '{contributor}' is not of type Person"
        self.contributor = contributor
    
    def getCreators(self) -> list[Person]:
        """Get creators list.

        Returns
        -------
        list[Person]
            A list of Person objects.
        """
        return self.creators

    def setCreators(self, creators:list[Person]):
        """Set the creators list.

        Parameters
        ----------
        creators : list[Person]
            The list to assign.
        """
        assert isinstance(creators, list), "Error: creators is not a list"
        assert all(isinstance(x, Person) for x in creators), f"Error: '{creators}' is not of type Person"
        self.creators = creators

    def addCreator(self, creators:Person):
        """Add a value to the creators list.

        Parameters
        ----------
        creators : Person
            The value to append.
        """
        assert isinstance(creators, Person), f"Error: '{creators}' is not of type Person"
        self.creators.append(creators)
        
    def getDataSource(self) -> str:
        """Get dataSource.

        Returns
        -------
        str
            The current value of dataSource.
        """
        return self.dataSource

    def setDataSource(self, dataSource:str):
        """Set dataSource.

        Parameters
        ----------
        dataSource : str
            The value to assign.
        """
        assert isinstance(dataSource, str), f"Error: '{dataSource}' is not of type str"
        self.dataSource = dataSource
    
    def getDatasetId(self) -> str:
        """Get datasetId.

        Returns
        -------
        str
            The current value of datasetId.
        """
        return self.datasetId

    def setDatasetId(self, datasetId:str):
        """Set datasetId.

        Parameters
        ----------
        datasetId : str
            The value to assign.
        """
        assert isinstance(datasetId, str), f"Error: '{datasetId}' is not of type str"
        self.datasetId = datasetId
    
    def getFundings(self) -> list[Funding]:
        """Get fundings list.

        Returns
        -------
        list[Funding]
            A list of Funding objects.
        """
        return self.fundings

    def setFundings(self, fundings:list[Funding]):
        """Set the fundings list.

        Parameters
        ----------
        fundings : list[Funding]
            The list to assign.
        """
        assert isinstance(fundings, list), "Error: fundings is not a list"
        assert all(isinstance(x, Funding) for x in fundings), f"Error: '{fundings}' is not of type Funding"
        self.fundings = fundings

    def addFunding(self, fundings:Funding):
        """Add a value to the fundings list.

        Parameters
        ----------
        fundings : Funding
            The value to append.
        """
        assert isinstance(fundings, Funding), f"Error: '{fundings}' is not of type Funding"
        self.fundings.append(fundings)
        
    def getInvestigators(self) -> list[Person]:
        """Get investigators list.

        Returns
        -------
        list[Person]
            A list of Person objects.
        """
        return self.investigators

    def setInvestigators(self, investigators:list[Person]):
        """Set the investigators list.

        Parameters
        ----------
        investigators : list[Person]
            The list to assign.
        """
        assert isinstance(investigators, list), "Error: investigators is not a list"
        assert all(isinstance(x, Person) for x in investigators), f"Error: '{investigators}' is not of type Person"
        self.investigators = investigators

    def addInvestigator(self, investigators:Person):
        """Add a value to the investigators list.

        Parameters
        ----------
        investigators : Person
            The value to append.
        """
        assert isinstance(investigators, Person), f"Error: '{investigators}' is not of type Person"
        self.investigators.append(investigators)
        
    def getLocation(self) -> Location:
        """Get location.

        Returns
        -------
        Location
            The current value of location.
        """
        return self.location

    def setLocation(self, location:Location):
        """Set location.

        Parameters
        ----------
        location : Location
            The value to assign.
        """
        assert isinstance(location, Location), f"Error: '{location}' is not of type Location"
        self.location = location
    
    def getName(self) -> str:
        """Get name.

        Returns
        -------
        str
            The current value of name.
        """
        return self.name

    def setName(self, name:str):
        """Set name.

        Parameters
        ----------
        name : str
            The value to assign.
        """
        assert isinstance(name, str), f"Error: '{name}' is not of type str"
        self.name = name
        self.id = self.ns + '/' + name # This is a hack to set the id of the dataset based on the name

    def getNotes(self) -> str:
        """Get notes.

        Returns
        -------
        str
            The current value of notes.
        """
        return self.notes

    def setNotes(self, notes:str):
        """Set notes.

        Parameters
        ----------
        notes : str
            The value to assign.
        """
        assert isinstance(notes, str), f"Error: '{notes}' is not of type str"
        self.notes = notes
    
    def getOriginalDataUrl(self) -> str:
        """Get originalDataUrl.

        Returns
        -------
        str
            The current value of originalDataUrl.
        """
        return self.originalDataUrl

    def setOriginalDataUrl(self, originalDataUrl:str):
        """Set originalDataUrl.

        Parameters
        ----------
        originalDataUrl : str
            The value to assign.
        """
        assert isinstance(originalDataUrl, str), f"Error: '{originalDataUrl}' is not of type str"
        self.originalDataUrl = originalDataUrl
    
    def getPaleoData(self) -> list[PaleoData]:
        """Get paleoData list.

        Returns
        -------
        list[PaleoData]
            A list of PaleoData objects.
        """
        return self.paleoData

    def setPaleoData(self, paleoData:list[PaleoData]):
        """Set the paleoData list.

        Parameters
        ----------
        paleoData : list[PaleoData]
            The list to assign.
        """
        assert isinstance(paleoData, list), "Error: paleoData is not a list"
        assert all(isinstance(x, PaleoData) for x in paleoData), f"Error: '{paleoData}' is not of type PaleoData"
        self.paleoData = paleoData

    def addPaleoData(self, paleoData:PaleoData):
        """Add a value to the paleoData list.

        Parameters
        ----------
        paleoData : PaleoData
            The value to append.
        """
        assert isinstance(paleoData, PaleoData), f"Error: '{paleoData}' is not of type PaleoData"
        self.paleoData.append(paleoData)
        
    def getPublications(self) -> list[Publication]:
        """Get publications list.

        Returns
        -------
        list[Publication]
            A list of Publication objects.
        """
        return self.publications

    def setPublications(self, publications:list[Publication]):
        """Set the publications list.

        Parameters
        ----------
        publications : list[Publication]
            The list to assign.
        """
        assert isinstance(publications, list), "Error: publications is not a list"
        assert all(isinstance(x, Publication) for x in publications), f"Error: '{publications}' is not of type Publication"
        self.publications = publications

    def addPublication(self, publications:Publication):
        """Add a value to the publications list.

        Parameters
        ----------
        publications : Publication
            The value to append.
        """
        assert isinstance(publications, Publication), f"Error: '{publications}' is not of type Publication"
        self.publications.append(publications)
        
    def getSpreadsheetLink(self) -> str:
        """Get spreadsheetLink.

        Returns
        -------
        str
            The current value of spreadsheetLink.
        """
        return self.spreadsheetLink

    def setSpreadsheetLink(self, spreadsheetLink:str):
        """Set spreadsheetLink.

        Parameters
        ----------
        spreadsheetLink : str
            The value to assign.
        """
        assert isinstance(spreadsheetLink, str), f"Error: '{spreadsheetLink}' is not of type str"
        self.spreadsheetLink = spreadsheetLink
    
    def getVersion(self) -> str:
        """Get version.

        Returns
        -------
        str
            The current value of version.
        """
        return self.version

    def setVersion(self, version:str):
        """Set version.

        Parameters
        ----------
        version : str
            The value to assign.
        """
        assert isinstance(version, str), f"Error: '{version}' is not of type str"
        self.version = version
    


BLACKLIST = {
    'metadataMD5' : 1,
    'paleoData_paleoDataMD5' : 1,
    'paleoData_paleoMeasurementTableMD5' : 1,
    'paleoDataMD5' : 1,
    'paleoMeasurementTableMD5' : 1,
    'tagMD5' : 1,
    'chronData_chronDataMD5' : 1,
    'chronData_chronMeasurementTableMD5' : 1,
    'chronDataMD5' : 1,
    'chronMeasurementTableMD5' : 1
}

REVERSE_BLACKLIST = {
    'inferredFrom' : 1,
    'foundInTable' : 1,
    'foundInDataset' : 1
    #'takenAtDepth' : 1
}


"""
The LiPD class describes a `LiPD (Linked Paleo Data) <https://cp.copernicus.org/articles/12/1093/2016/cp-12-1093-2016.html>`_ object. It contains an `RDF <https://www.w3.org/RDF/>`_ Graph which is serialization of the LiPD data into an RDF graph containing terms from the `LiPD Ontology <http://linked.earth/Ontology/release/core/1.2.0/index-en.html>`
How to browse and query LiPD objects is described in a short example below, while `this notebook <https://nbviewer.jupyter.org/github/LinkedEarth/pylipd/blob/master/example_notebooks/pylipd_tutorial.ipynb>`_ demonstrates how to use PyLiPD to view and query LiPD datasets.
"""

import ast
import os
import re
import os.path
import tempfile
import pandas as pd
import random
import string
import io
import numpy as np
import json
import uuid

from pylipd.classes.dataset import Dataset

from pylipd.utils.json_to_rdf import JSONToRDF
from pylipd.utils.rdf_to_json import RDFToJSON

from .globals.queries import QUERY_FILTER_TIME, QUERY_BIBLIO, QUERY_DSID, QUERY_DSNAME, QUERY_ENSEMBLE_TABLE, QUERY_ENSEMBLE_TABLE_SHORT, QUERY_FILTER_ARCHIVE_TYPE, QUERY_FILTER_GEO, QUERY_VARIABLE, QUERY_VARIABLE_GRAPH, QUERY_UNIQUE_ARCHIVE_TYPE, QUERY_TIMESERIES_ESSENTIALS_CHRON, QUERY_TIMESERIES_ESSENTIALS_PALEO, QUERY_DISTINCT_VARIABLE, QUERY_DATASET_PROPERTIES, QUERY_VARIABLE_PROPERTIES, QUERY_MODEL_PROPERTIES, QUERY_LOCATION, QUERY_FILTER_DATASET_NAME, QUERY_FILTER_COMPILATION, QUERY_COMPILATION_NAME

from .lipd_series import LiPDSeries
from .utils.multi_processing import multi_convert_to_rdf, multi_load_lipd
from .utils.rdf_graph import RDFGraph

from .utils.rdf_to_lipd import RDFToLiPD
from .utils.legacy_utils import LiPD_Legacy
from .utils.utils import sanitizeId

#import bibtexparser
#from bibtexparser.bibdatabase import BibDatabase
from doi2bib import crossref
from pybtex.database import BibliographyData, Entry

from .globals.urls import NSURL, DEFAULT_GRAPH_URI

class LiPD(RDFGraph):
    '''The LiPD class describes a `LiPD (Linked Paleo Data) <https://cp.copernicus.org/articles/12/1093/2016/cp-12-1093-2016.html>`_ object. It contains an `RDF <https://www.w3.org/RDF/>`_ Graph which is serialization of the LiPD data into an RDF graph containing terms from the `LiPD Ontology <http://linked.earth/Ontology/release/core/1.2.0/index-en.html>`_.
    How to browse and query LiPD objects is described in a short example below.

    Examples
    --------
    In this example, we read an online LiPD file and convert it into a time series object dictionary.

    .. jupyter-execute::
        
        from pylipd.lipd import LiPD

        lipd = LiPD()        
        lipd.load(["https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1//Nunalleq.Ledger.2018.lpd"])
        
        ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())

        for dsname, tsos in ts_list.items():
            for tso in tsos:
                if 'paleoData_variableName' in tso:
                    print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])
    '''
    def __init__(self, graph=None):
        super().__init__(graph)

    
    def load_from_dir(self, dir_path, parallel=False, cutoff=None, standardize=True, add_labels=True):
        '''Load LiPD files from a directory
       
        Parameters
        ----------

        dir_path : str
            path to the directory containing lipd files

        parallel: bool
            (Optional) set to True to process lipd files in parallel. You *must* run this function under the "__main__" process for this to work
        
        cutoff : int
            (Optional) the maximum number of files to load at once.
            
        Examples
        --------
        In this example, we load LiPD files from a directory.

        .. jupyter-execute::
            
            from pylipd.lipd import LiPD

            lipd = LiPD()        
            lipd.load_from_dir("../examples/data")

            print(lipd.get_all_dataset_names())
        '''
        if not os.path.isdir(dir_path):
            print(f"Directory {dir_path} does not exist")
            return

        lipdfiles = []
        for path in os.listdir(dir_path):
            file_path = os.path.join(dir_path, path)
            if os.path.isfile(file_path) and path.endswith(".lpd"):
                lipdfiles.append(file_path)
        if cutoff:
            lipdfiles = lipdfiles[0:cutoff]
        self.load(lipdfiles, parallel, standardize, add_labels)


    # Allows loading http locations
    def load(self, lipdfiles, parallel=False, standardize=True, add_labels=True):
        '''Load LiPD files. 
        

        Parameters
        ----------

        lipdfiles : list of str
            array of paths to lipd files (the paths could also be urls)

        parallel: bool
            (Optional) set to True to process lipd files in parallel. You *must* run this function under the "__main__" process for this to work


        Examples
        --------
        In this example, we load LiPD files for an array of paths.

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            lipd = LiPD() 
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd",
                "../examples/data/Ant-WAIS-Divide.Severinghaus.2012.lpd",
                "https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1/Nunalleq.Ledger.2018.lpd"                    
            ])            

            print(lipd.get_all_dataset_names())
        '''        
        if type(lipdfiles) is not list:
            lipdfiles = [lipdfiles]
            
        numfiles = len(lipdfiles)
        print(f"Loading {numfiles} LiPD files")
        self.graph = multi_load_lipd(self.graph, lipdfiles, parallel, standardize, add_labels)
        print("Loaded..")
    
    #def load_from_lipdverse(self, datasetID, version=None):
        


    def convert_lipd_dir_to_rdf(self, lipd_dir, rdf_file, parallel=False, standardize=True, add_labels=False):
        '''Convert a directory containing LiPD files into a single RDF file (to be used for uploading to Knowledge Bases like GraphDB)

        Parameters
        ----------

        lipd_dir : str
            Path to the directory containing lipd files

        rdf_file : str
            Path to the output rdf file

        '''

        filemap = {}
        for path in os.listdir(lipd_dir):
            fullpath = os.path.join(lipd_dir, path)
            tmp_rdf_file = tempfile.NamedTemporaryFile().name
            filemap[fullpath] = tmp_rdf_file
        
        print(f"Converting {len(filemap.keys())} LiPD files to RDF..")

        multi_convert_to_rdf(filemap, parallel, standardize, add_labels)
        
        print("Conversion to RDF done..")

        print("Writing to main RDF file..")
        with open(rdf_file, "w") as fout:
            for lipdfile in filemap.keys():
                tmp_rdf_file = filemap[lipdfile]
                if os.path.exists(tmp_rdf_file):
                    fin = open(tmp_rdf_file, "r")
                    data = fin.read();
                    fin.close()
                    fout.write(data)
                    os.remove(tmp_rdf_file)
            fout.close()
        print("Written..")


    def load_remote_datasets(self, dsnames, load_default_graph=True):
        '''Loads remote datasets into cache if a remote endpoint is set

        Parameters
        ----------

        dsnames : array
            array of dataset names

        Examples
        --------

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Fetch LiPD data from remote RDF Graph
            lipd_remote = LiPD()
            lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse-dynamic")
            lipd_remote.load_remote_datasets(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"])
            print(lipd_remote.get_all_dataset_names())
        '''
        if not self.endpoint:
            raise Exception("No remote endpoint")
        
        if type(dsnames) is not list:
            dsnames = [dsnames]
            
        if dsnames == None or len(dsnames) == 0:
            raise Exception("No dataset names to cache")
        dsnamestr = (' '.join('<' + NSURL + "/" + dsname + '>' for dsname in dsnames))

        if load_default_graph:
            dsnamestr += f" <{DEFAULT_GRAPH_URI}>"
        print("Caching datasets from remote endpoint..")
        qres, qres_df = self.query(f"SELECT ?s ?p ?o ?g WHERE {{ GRAPH ?g {{ ?s ?p ?o }} VALUES ?g {{ {dsnamestr} }} }}", remote=True)

        # Reinitialize graph
        # self._initialize_graph()
        for row in qres:
            self.graph.add((row.s, row.p, row.o, row.g))
        print("Done..")
    

    def update_remote_datasets(self, dsnames):
        '''Updates local LiPD Graph for datasets to remote endpoint'''
        if not self.endpoint:
            raise Exception("No remote endpoint")
        # TODO: Implement this


    def get_bibtex(self, remote = True, save = True, path = 'mybiblio.bib', verbose = False):
        '''Get BibTeX for loaded datasets
        
        Parameters
        ----------
        remote : bool 
            (Optional) If set to True, will return the bibliography by checking against the DOI
        
        save : bool
            (Optional) Whether to save the bibliography to a file
            
        path : str
            (Optional) Path where to save the file
        
        verbose : bool
            (Optional) Whether to print out on the console. Note that this option will turn on automatically if saving to a file fails. 

        Returns
        -------
        bibs : list
            List of BiBTex entry
        
        df : pandas.DataFrame
            Bibliography information in a Pandas DataFrame
        
        Examples
        --------    
        
        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Fetch LiPD data from remote RDF Graph
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            print(lipd.get_bibtex(save=False))
        '''

        def establish_type(pub_type):
            
            if pub_type:
                pub_type = re.sub('-', '', pub_type).lower()
            else:
                pub_type = 'misc'
            
            if re.match(r".*article.*", pub_type) or re.match(r".*shortcommunication.*", pub_type):
                pub_type = 'article'
            elif re.match(r".*chapter.*", pub_type) or re.match(r".*book.*", pub_type):
                pub_type = 'chapter'
            elif re.match(r".*report.*", pub_type):
                pub_type = 'report'
            else:
                pub_type = 'misc'
            
            return pub_type

        def make_bib(row):
            pub_type = establish_type(row['type'])
            
            # Create a unique citation ID if not given
            row = row.fillna("")
            if row['citeKey'] is None:
                characters = string.ascii_letters + string.digits
                citation_key = ''.join(random.choice(characters) for i in range(8))
            else:
                citation_key = row['citeKey']
            
            entries = [] #start creating the list
            
            if row['authors']:
                entries.append(('author', str(row['authors'])))
            if row['doi']:
                entries.append(('doi',str(row['doi'])))
            if row['year']:
                entries.append(('year',str(row['year'])))
            if row['pubyear']:
                entries.append(('year',str(row['pubyear'])))
            if row['title']:
                if pub_type == 'article' or pub_type == 'misc':
                    entries.append(('title',str(row['title'])))
                elif pub_type == 'chapter' or pub_type == 'report':
                    entries.append(('chapter', str(row['title'])))
            if row['journal']:
                if pub_type == 'article':
                    entries.append(('journal', str(row['journal'])))
                if pub_type == 'book':
                    entries.append(('title', str(row['journal'])))
            if row['volume']:
                entries.append(('volume', str(row['volume'])))
            if row['issue']:
                entries.append(('issue', str(row['issue'])))
            if row['pages']:
                entries.append(('pages', str(row['pages'])))
            if row['publisher']:
                entries.append(('publisher', str(row['publisher'])))
            if row['report']:
                entries.append(('title', str(row['report'])))
            if row['edition']:
                entries.append(('edition', str(row['edition'])))
            if row['institution']:
                entries.append(('institution',str(row['institution'])))
            if row['url']:
                entries.append(('url',str(row['url'])))
            if row['url2']:
                entries.append(('url',str(row['url2'])))
            
            if pub_type == 'article':
                bib = BibliographyData({citation_key:Entry('article',entries)})
            elif pub_type == 'chapter' or pub_type == 'report':
                bib = BibliographyData({citation_key:Entry('inbook',entries)})
            elif pub_type == 'misc':
                bib = BibliographyData({citation_key:Entry('misc',entries)})
            
            return bib  
        
        result, df = self.query(QUERY_BIBLIO)
        
        bibs = []

        for idx,row in df.iterrows():
            if remote == True:
                try: 
                    f = (crossref.get_bib(row['doi']))
                    if f[0]==True:
                        bibs.append(f[1])
                    else:
                        print(f"Cannot find a matching record for the provided DOI ({row['doi']}), creating the entry manually")
                        bibs.append(make_bib(row).to_string('bibtex'))
                except:
                    print("Cannot parse the provided DOI, creating the entry manually")
                    bibs.append(make_bib(row).to_string('bibtex'))
            
        if save == True:   
            try:
                with io.open(path, 'w', encoding="utf-8") as bibfile:
                    for bib in bibs:
                        bibfile.write("{}\n".format(bib))

            except TypeError:
                print("Can't save in output file\n")
                verbose = True
        
        if verbose == True:
            print(bibs)
        
        return bibs, df       

    def get_timeseries(self, dsnames, to_dataframe=False, mode="paleo", time="age"):
        '''Get Legacy LiPD like Time Series Object (tso)
        
        This function is meant to provide legacy support to the older version of the LiPD utilities, which returns a dictionary of timeseries objects. The function also supports returning to a pandas.DataFrame, essentially flattening all the information. This is useful to explore all possible properties but can be slow for large number of datasets or if you only require some standard information. In this case, use `get_timeseries_essentials`. 

        Parameters
        ----------

        dsnames : list
            array of dataset id or name strings
        
        to_dataframe : bool {True; False}
            Whether to return a dataframe along the dictionary. Default is False
        
        mode: 'paleo' or 'chron'
            Whether to return information from the PaleoData or ChronData objects
        
        time: 'age' or 'year'
            Whether the time is expressed as year or age
        
        Returns
        -------
        
        ts : dict
            A dictionary containing Time Series Object
            
        df : Pandas.DataFrame
            If to_dataframe is set to True, returns a queryable Pandas DataFrame

        Examples
        --------
        
        To only return a list of timeseries objects

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Fetch LiPD data from remote RDF Graph
            lipd_remote = LiPD()
            lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse-dynamic")
            ts_list = lipd_remote.get_timeseries(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"])
            for dsname, tsos in ts_list.items():
                for tso in tsos:
                    if 'paleoData_variableName' in tso:
                        print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])
        
        
        To return a dataframe in addition to the list of timeseries objects
        
        
        .. jupyter-execute::
            
            from pylipd.lipd import LiPD
            
            lipd_remote = LiPD()
            lipd_remote.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse-dynamic")
            ts_list, df = lipd_remote.get_timeseries(["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001", "MD98_2181.Stott.2007", "Ant-WAIS-Divide.Severinghaus.2012"], to_dataframe = True)
            df.head()
        
        See also
        --------
        
        pylipd.lipd.LiPD.get_timeseries_essentials
        
        '''
        
        if type(dsnames)==str:
            dsnames=[dsnames]
        
        ts = self._get_timeseries(dsnames, mode=mode, time=time)
        if to_dataframe == False:
            return ts
        elif to_dataframe == True:
            dict_list =[]

            for item in ts.keys():
                for dictionary in ts[item]:
                    dict_list.append(dictionary)

            df = pd.DataFrame.from_dict(dict_list, orient='columns')
            
            return ts, df

    def _get_timeseries(self, dsnames, mode="paleo", time="age"):
        timeseries = {}
        for dsname in dsnames:
            converter = RDFToLiPD(self.graph)
            d = converter.convert_to_json(dsname)
            print("Extracting timeseries from dataset: " + dsname + " ...")
            if len(d.items()):
                tss = LiPD_Legacy().extract(d, mode=mode, time=time)
                timeseries[dsname] = tss
        return timeseries
    
    def get_timeseries_essentials(self, dsnames=None, mode='paleo'):
        ''' Returns specific properties for timeseries: 'dataSetName', 'archiveType', 'geo_meanLat', 'geo_meanLon',
               'geo_meanElev', 'paleoData_variableName', 'paleoData_values',
               'paleoData_units', 'paleoData_proxy' (paleo only), 'paleoData_proxyGeneral' (paleo only),
               'time_variableName', 'time_values', 'time_units', 'depth_variableName',
               'depth_values', 'depth_units'
        

        Parameters
        ----------
        dsnames : list
            array of dataset id or name strings        
        mode : paleo, chron
            Whether to return the information stored in the PaleoMeasurementTable or the ChronMeasurementTable. The default is 'paleo'.

        Raises
        ------
        ValueError
            Need to select either 'chron' or 'paleo'

        Returns
        -------
        qres_df : pandas.DataFrame
            A pandas dataframe returning the properties in columns for each series stored in a row of the dataframe

        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_datasets
            lipd = load_datasets('ODP846.Lawrence.2006.lpd')
            df_paleo = lipd.get_timeseries_essentials(mode='paleo')
            print(df_paleo)
        
        To return the information stored in the ChronTable:
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_datasets
            lipd = load_datasets('ODP846.Lawrence.2006.lpd')  
            df_chron = lipd.get_timeseries_essentials(mode='chron')
            print(df_chron)
    
        '''
        
        if dsnames is None:
            dsnames= ''
        if type(dsnames)==str:
            dsnames=[dsnames]            
        
        qres_df = None
        for dsname in dsnames:
            if mode == 'paleo':
                query = QUERY_TIMESERIES_ESSENTIALS_PALEO
                query = query.replace("[dsname]", dsname)
            elif mode == 'chron':
                query = QUERY_TIMESERIES_ESSENTIALS_CHRON
                query = query.replace("[dsname]", dsname)
            else:
                raise ValueError("The mode should be either 'paleo' or 'chron'")
        
            qres, qtmp_df = self.query(query)
            
            try:
                qtmp_df['paleoData_values']=qtmp_df['paleoData_values'].apply(lambda row : np.array(json.loads(row)))
            except:
                qtmp_df['chronData_values']=qtmp_df['chronData_values'].apply(lambda row : np.array(json.loads(row)))
            
            
            qtmp_df['time_values']=qtmp_df['time_values'].apply(lambda x : np.array(json.loads(x)) if x is not None else None)
            qtmp_df['depth_values']=qtmp_df['depth_values'].apply(lambda x : np.array(json.loads(x)) if x is not None else None)
            if qres_df is None:
                qres_df = qtmp_df
            else:
                qres_df = pd.concat([qres_df, qtmp_df], ignore_index=True)
        
        
        return qres_df
            

    def get_lipd(self, dsname):
        '''Get LiPD json for a dataset

        Parameters
        ----------

        dsname : str
            dataset id

        Returns
        -------

        lipdjson : dict
            LiPD json

        Examples
        --------

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Load a local LiPD file
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
            ])
            lipd_json = lipd.get_lipd(lipd.get_all_dataset_names()[0])
            print(lipd_json)
        '''           
        converter = RDFToLiPD(self.graph)
        return converter.convert_to_json(dsname)

    def create_lipd(self, dsname, lipdfile):
        '''Create LiPD file for a dataset

        Parameters
        ----------

        dsname : str
            dataset id

        lipdfile: str
            path to LiPD file

        Returns
        -------

        lipdjson : dict
            LiPD json

        Examples
        --------

        .. jupyter-execute::


            from pylipd.lipd import LiPD

            # Load a local file
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
            ])
            dsname = lipd.get_all_dataset_names()[0]
            lipd.create_lipd(dsname, "test.lpd")
        '''           
        converter = RDFToLiPD(self.graph)
        return converter.convert(dsname, lipdfile)
    

    def get(self, dsnames):
        '''Gets dataset(s) from the graph and returns the popped LiPD object
        
        Parameters
        ----------
        dsnames : str or list of str
            dataset name(s) to get.
        
        Returns
        -------

        pylipd.lipd.LiPD
            LiPD object with the retrieved dataset(s)

        Examples
        --------
        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Load LiPD files from a local directory
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])

            all_datasets = lipd.get_all_dataset_names()
            print("Loaded datasets: " + str(all_datasets))
            ds = lipd.get(all_datasets[0])
            print("Got dataset: " + str(ds.get_all_dataset_names()))       
        '''
        dsnames = [dsnames] if type(dsnames) is not list else dsnames
        dsids = [(f"{NSURL}/{dsname}" if not dsname.startswith(NSURL) else dsname) for dsname in dsnames]

        ds = super().get(dsids)
        return LiPD(ds.graph)

    def pop(self, dsnames):
        '''Pops dataset(s) from the graph and returns the popped LiPD object
        
        Parameters
        ----------
        dsnames : str or list of str
            dataset name(s) to be popped.
        
        Returns
        -------

        pylipd.lipd.LiPD
            LiPD object with the popped dataset(s)

        Examples
        --------
        .. jupyter-execute::


            from pylipd.lipd import LiPD

            # Load local files
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            all_datasets = lipd.get_all_dataset_names()
            print("Loaded datasets: " + str(all_datasets))
            popped = lipd.pop(all_datasets[0])
            print("Loaded datasets after pop: " + str(lipd.get_all_dataset_names()))
            print("Popped dataset: " + str(popped.get_all_dataset_names()))       
        '''

        dsnames = [dsnames] if type(dsnames) is not list else dsnames
        dsids = [(f"{NSURL}/{dsname}" if not dsname.startswith(NSURL) else dsname) for dsname in dsnames]
        popped = super().pop(dsids)
        return LiPD(popped.graph)

    def remove(self, dsnames):
        '''Removes dataset(s) from the graph
        
        Parameters
        ----------
        dsnames : str or list of str
            dataset name(s) to be removed
        
        Examples
        --------
        .. jupyter-execute::


            from pylipd.lipd import LiPD
            
            # Load local files
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            all_datasets = lipd.get_all_dataset_names()
            print("Loaded datasets: " + str(all_datasets))
            lipd.remove(all_datasets[0])
            print("Loaded datasets after remove: " + str(lipd.get_all_dataset_names()))
        '''
        
        dsnames = [dsnames] if type(dsnames) is not list else dsnames
        dsids = [(f"{NSURL}/{dsname}" if not dsname.startswith(NSURL) else dsname) for dsname in dsnames]

        super().remove(dsids)


    def get_all_dataset_names(self):
        '''Get all Dataset Names
        
        Returns
        -------
        
        dsnames : list
        
        A list of datasetnames
        
        Examples
        --------

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Load local files
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            print(lipd.get_all_dataset_names())
        '''        
        qres, qres_df = self.query(QUERY_DSNAME)
        return [sanitizeId(row.dsname) for row in qres]

    def get_all_dataset_ids(self):
        '''Get all Dataset ids
        
        Returns
        -------
        
        dsids : list
        
        A list of datasetnames
        
        Examples
        --------

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Load local files
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            print(lipd.get_all_dataset_ids())
        '''
        qres, qres_df = self.query(QUERY_DSID)
        return [sanitizeId(row.dsid) for row in qres]
    
    def get_all_archiveTypes(self):
        '''
        Returns a list of all the unique archiveTypes present in the LiPD object

        Returns
        -------
        list
            A list of archiveTypes
            
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.lipd import LiPD

            # Load Local files
            lipd = LiPD()
            lipd.load([
                "../examples/data/Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001.lpd",
                "../examples/data/MD98_2181.Stott.2007.lpd"
            ])
            print(lipd.get_all_archiveTypes())

        '''
        
        qres, qres_df = self.query(QUERY_UNIQUE_ARCHIVE_TYPE)
        return [str(row.archiveType) for row in qres]
        
    def get_all_locations(self, dsname = None):
        '''Return geographical coordinates for all the datasets.       

        Parameters
        ----------
        dsname : str, optional
            The name of the dataset for which to return the timeseries information. The default is None.

        Returns
        -------
        df : pandas.DataFrame
            A pandas dataframe returning the latitude, longitude and elevation for each dataset
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            df = lipd.get_all_locations()
            print(df)

        '''
        
        if dsname is None:
            dsname= ''
        
        query = QUERY_LOCATION
        query = query.replace("[dsname]", dsname)
            
                
        return self.query(query)[1]
    
    def get_all_compilation_names(self):
        '''Return the names of the compilation present in the LiPD object     

        Returns
        -------
        l : list
            A list returning the names of the available compilations.
        
        Examples
        --------
        
        .. jupyter-execute::
        
            from pylipd.utils.dataset import load_dir
            
            lipd = load_dir('Temp12k')
            df = lipd.get_all_compilation_names()
            print(df)

        '''
        
        qres, qres_df = self.query(QUERY_COMPILATION_NAME)
        return [sanitizeId(row.compilationName) for row in qres]


    def get_ensemble_tables(self, dsname = None, ensembleVarName = None, ensembleDepthVarName = 'depth'):
        '''Gets ensemble tables from the LiPD graph

        Parameters
        ----------

        dsname : str
            The name of the dataset if you wish to analyse one at a time (Set to ".*" to match all datasets with a common root)
        
        ensembleVarName : None or str
            ensemble variable name. Default is None, which searches for names that contain "year" or "age" (Set to ".*" to match all ensemble variable names)
        
        ensembleDepthVarName : str
            ensemble depth variable name. Default is 'depth' (Set to ".*" to match all ensemble depth variable names)

        Returns
        -------

        ensemble_tables : dataframe
            A dataframe containing the ensemble tables


        Examples
        --------

        .. jupyter-execute::

            from pylipd.lipd import LiPD

            lipd = LiPD()
            lipd.load([
                "../examples/data/ODP846.Lawrence.2006.lpd"
            ])
            all_datasets = lipd.get_all_dataset_names()
            print("Loaded datasets: " + str(all_datasets))

            ens_df = lipd.get_ensemble_tables(
                ensembleVarName="age",
                ensembleDepthVarName="depth"
            )
            print(ens_df)
        '''
        
        if dsname is None:
            dsname = ''
        
        if ensembleVarName is None:
            query = QUERY_ENSEMBLE_TABLE_SHORT
            query = query.replace("[dsname]", dsname)
            query = query.replace("[ensembleDepthVarName]", ensembleDepthVarName)
        
        else:
       
            query = QUERY_ENSEMBLE_TABLE
            query = query.replace("[dsname]", dsname)
            query = query.replace("[ensembleVarName]", ensembleVarName)
            query = query.replace("[ensembleDepthVarName]", ensembleDepthVarName)

        qres, qres_df = self.query(query)

        nan_replace = re.compile(re.escape('NaN'), re.IGNORECASE)

        qres_df['ensembleDepthValues']=qres_df['ensembleDepthValues'].apply(lambda row : np.array(json.loads(row)))
        qres_df['ensembleVariableValues']=qres_df['ensembleVariableValues'].apply(lambda row : np.array(ast.literal_eval(nan_replace.sub('None', row))))
        
        return qres_df


    def get_all_variables(self):
        '''
        Returns a list of all variables in the graph
        
        Returns
        -------

        pandas.DataFrame
            A dataframe of all variables in the graph with columns uri, varid, varname
            
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.lipd import LiPD

            lipd = LiPD()
            lipd.load([
                "../examples/data/ODP846.Lawrence.2006.lpd"
            ])
            
            df = lipd.get_all_variables()
            print(df)

        '''
        return self.query(QUERY_VARIABLE)[1]

    def get_all_variable_names(self):
        """
        Get a list of all possible distinct variableNames. Useful for filtering and querying. 

        Returns
        -------
        list
            A list of unique variableName 
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            varName = lipd.get_all_variable_names()
            print(varName)
        

        """
        
        return self.query(QUERY_DISTINCT_VARIABLE)[1].iloc[:,0].values.tolist()
    
    def get_dataset_properties(self):
        """Get a list of unique properties attached to a dataset. 
        
        Note: Some properties will return another object (e.g., 'publishedIn' will give you a Publication object with its own properties)
        Note: Not all datasets will have the same available properties (i.e., not filled in by a user)
        

        Returns
        -------
        clean_list : list
            A list of available properties that can queried

        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir(name='Pages2k')
            dataset_properties = lipd.get_dataset_properties()
            print(dataset_properties)
        """
        
        query_list = self.query(QUERY_DATASET_PROPERTIES)[1].iloc[:,0].values.tolist()
        clean_list = [item.split("#")[-1] for item in query_list]
        
        return clean_list
    
    def get_variable_properties(self):
        '''Get a list of variable properties that can be used for querying
        

        Returns
        -------
        list
            A list of unique variable properties
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir(name='Pages2k')
            variable_properties = lipd.get_variable_properties()
            print(variable_properties)

        '''
        
        query_list = self.query(QUERY_VARIABLE_PROPERTIES)[1].iloc[:,0].values.tolist()
        clean_list = [item.split("#")[-1] for item in query_list]
        
        return clean_list
    
    def get_model_properties(self):
        '''Get all the properties associated with a model
        

        Returns
        -------
        List
            A list of unique properties attached to models
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_datasets
            lipd = load_datasets(names='ODP846')
            model_properties = lipd.get_model_properties()
            print(model_properties)


        '''
        
        query_list = self.query(QUERY_MODEL_PROPERTIES)[1].iloc[:,0].values.tolist()
        clean_list = [item.split("#")[-1] for item in query_list]
        
        return clean_list

    def to_lipd_series(self, parallel=False):
        '''
        Converts the LiPD object to a LiPDSeries object

        Parameters
        ----------
        parallel : bool
            Whether to use parallel processing to load the data. Default is False

        Returns
        -------
        pylipd.lipd.LiPDSeries
            A LiPDSeries object
            
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.lipd import LiPD

            lipd = LiPD()
            lipd.load([
                "../examples/data/ODP846.Lawrence.2006.lpd"
            ])
            
            S = lipd.to_lipd_series()
        
        '''
        S = LiPDSeries()    
        S.load(self, parallel)
        return S


    # bbox = left,bottom,right,top
    # bbox = min Longitude , min Latitude , max Longitude , max Latitude 
    def filter_by_geo_bbox(self, lonMin, latMin, lonMax, latMax):
        '''
        Filters datasets to return a new LiPD object that only keeps datasets that fall within the bounding box

        Parameters
        ----------
        lonMin : float
            Minimum longitude

        latMin : float
            Minimum latitude
        
        lonMax : float
            Maximum longitude

        latMax : float
            Maximum latitude

        Returns
        -------

        pylipd.lipd.LiPD
            A new LiPD object that only contains datasets that fall within the bounding box
        
        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir()
            Lfiltered = lipd.filter_by_geo_bbox(0,25,50,50)
            Lfiltered.get_all_dataset_names()           
        
        '''
        query = QUERY_FILTER_GEO
        query = query.replace("[lonMin]", str(lonMin))
        query = query.replace("[latMin]", str(latMin))
        query = query.replace("[lonMax]", str(lonMax))               
        query = query.replace("[latMax]", str(latMax))
        qres, qres_df = self.query(query)
        dsnames = [sanitizeId(row.dsname) for row in qres]
        return self.get(dsnames)


    def filter_by_archive_type(self, archiveType):
        '''
        Filters datasets to return a new LiPD object that only keeps datasets that have the specified archive type

        Parameters
        ----------

        archiveType : str
            The archive type to filter by

        Returns
        -------
        
        pylipd.lipd.LiPD
            A new LiPD object that only contains datasets that have the specified archive type (regex)
        
        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            Lfiltered = lipd.filter_by_archive_type('marine')
            Lfiltered.get_all_archiveTypes()
        
        If searching for multiple archiveTypes, you can construct the name as follows:
        
        .. jupyter-execute::

            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            Lfiltered = lipd.filter_by_archive_type('marine|coral')
            Lfiltered.get_all_archiveTypes()
        
        '''
        
        query = QUERY_FILTER_ARCHIVE_TYPE
        query = query.replace("[archiveType]", archiveType)
        qres, qres_df = self.query(query)
        dsnames = [sanitizeId(row.dsname) for row in qres]
        return self.get(dsnames)

    
    def filter_by_datasetName(self, datasetName):
        '''
        Filters datasets to return a new LiPD object that only keeps datasets that have the specified names

        Parameters
        ----------

        datasetName : str 
            The datasetNames to filter by

        Returns
        -------
        
        pylipd.lipd.LiPD
            A new LiPD object that only contains datasets that have the specified archive type (regex)
        
        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            Lfiltered = lipd.filter_by_datasetName('Ocn-RedSea.Felis.2000')
            Lfiltered.get_all_dataset_names()
        
        If searching for multiple dataset names, you can construct the name as follows:
        
        .. jupyter-execute::
        
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            dsnames = ['Ocn-RedSea.Felis.2000','Ant-WAIS-Divide.Severinghaus.2012']
            dsquery = '|'.join(dsnames)
            Lfiltered = lipd.filter_by_datasetName(dsquery)
            Lfiltered.get_all_dataset_names()
            
        '''
        
        query = QUERY_FILTER_DATASET_NAME
        query = query.replace("[datasetName]", datasetName)
        qres, qres_df = self.query(query)
        dsnames = [sanitizeId(row.dsname) for row in qres]
        return self.get(dsnames)
    
    def filter_by_compilationName(self, compilationName):
        '''
        Filters datasets to return a new LiPD object that only keeps datasets that have the specific compilation

        Parameters
        ----------

        compilationName : str 
            The name of the compilation to filter by

        Returns
        -------
        
        pylipd.lipd.LiPD
            A new LiPD object that only contains datasets that have the specified archive type (regex)
        
        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import available_dataset_names, load_datasets

            dsList = available_dataset_names()
            D = load_datasets(dsList)
            Dfiltered = D.filter_by_compilationName('Temp12k')
            Dfiltered.get_all_dataset_names()
        
        '''
        
        query = QUERY_FILTER_COMPILATION
        query = query.replace("[compilationName]", compilationName)
        qres, qres_df = self.query(query)
        dsnames = [sanitizeId(row.dataSetName) for row in qres]
        return self.get(dsnames)
        
    
    def filter_by_time(self,timeBound, timeBoundType = 'any', recordLength = None):
        """
        Filter the records according to a specified time interval and the length of the record within that interval. Note that this function assumes that all records use the same time representation. 
        
        If you are unsure about the time representation, you may need to use `.get_timeseries_essentials`. 

        Parameters
        ----------
        timeBound : list
            Minimum and Maximum age value to search for.
        timeBoundType : str, optional
            The type of querying to perform. Possible values include: "any", "entire", and "entirely".
            - any: Overlap any portions of matching datasets (default)
            - entirely: are entirely overlapped by matching datasets
            - entire: overlap entire matching datasets but dataset can be shorter than the bounds
            The default is 'any'.
        recordLength : float, optional
            The minimum length the record needs to have while matching the ageBound criteria. The default is None.

        Raises
        ------
        ValueError
            timeBoundType must take the values in ["any", "entire", and "entirely"]

        Returns
        -------
        pylipd.lipd.LiPD
            A new LiPD object that only contains datasets that have the specified time interval
            
        Examples
        --------
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            Lfiltered = lipd.filter_by_time(timeBound=[0,1800])
            Lfiltered.get_all_dataset_names()

        """
        
        if timeBound and timeBound[0]>timeBound[1]:
                timeBound = [timeBound[1],timeBound[0]]

        timeBoundType=timeBoundType.lower()

        query = QUERY_FILTER_TIME
        __, df = self.query(query)
        if recordLength is None:
            if timeBoundType == 'entirely':
                filter_df = df[(df['minage'] <= timeBound[0]) & (df['maxage'] >= timeBound[1])]
            elif timeBoundType == 'entire':
                filter_df = df[(df['minage'] >= timeBound[0]) & (df['maxage'] <= timeBound[1])]
            elif timeBoundType == 'any':
                filter_df = df[(df['minage'] <= timeBound[1])]
            else:
                raise ValueError("timeBoundType must be in ['any', 'entirely','entire']")
        else:
            if timeBoundType == 'entirely':
                filter_df = df[(df['minage'] <= timeBound[0]) & (df['maxage'] >= timeBound[1]) & (np.abs(df['maxage']-df['minage'])>=recordLength)]
            elif timeBoundType == 'entire':
                filter_df = df[(df['minage'] >= timeBound[0]) & (df['maxage'] <= timeBound[1]) & (np.abs(df['maxage']-df['minage'])>=recordLength)]
            elif timeBoundType == 'any':
                filter_df = df[(df['minage'] <= timeBound[1]) & (np.abs(df['minage']-timeBound[1])>=recordLength)]
            else:
                raise ValueError("timeBoundType must be in ['any', 'entirely','entire']")
            
        dsnames = list(filter_df['dsname'])
        return self.get(dsnames)
    
    
    def get_datasets(self) -> 'list[Dataset]':
        '''
        Return datasets as instances of the Dataset class

        Parameters
        ----------

        Returns
        -------
        
        list of pylipd.classes.Dataset
            A list of Dataset objects
        
        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            lipd.get_datasets()
        
        '''        
        datasets: list[Dataset] = []
        for dsname in self.get_all_dataset_names():
            dsuri = NSURL + "/" + dsname
            r2j = RDFToJSON(dsuri, self.graph)
            data = json.loads(r2j.to_json())
            ds = Dataset.from_data(dsuri, data)
            datasets.append(ds)
        return datasets
    

    def _generate_unique_id(self, prefix='PYD'):
        # Generate a random UUID
        random_uuid = uuid.uuid4()  # Generates a random UUID.
        
        # Convert UUID format to the specific format we need
        # UUID is usually in the form '1e2a2846-2048-480b-9ec6-674daef472bd' so we slice and insert accordingly
        id_str = str(random_uuid)
        formatted_id = f"{prefix}-{id_str[:5]}-{id_str[9:13]}-{id_str[14:18]}-{id_str[19:23]}-{id_str[24:28]}"
        
        return formatted_id

    def _fix_missing_ids(self, ds: Dataset):
        # Assign variable ids if not present
        # Assign datatable csv file name if not present
        pd_counter = 0
        for pd in ds.getPaleoData():
            table_counter = 0
            for table in pd.getMeasurementTables():
                if not table.getFileName():
                    table.setFileName(f"paleo{pd_counter}measurement{table_counter}.csv")
                for v in table.getVariables():
                    if not v.getVariableId():
                        v.setVariableId(self._generate_unique_id(prefix='TS'))
                table_counter += 1
            pd_counter += 1

        chron_counter = 0
        for chron in ds.getChronData():
            table_counter = 0
            for table in chron.getMeasurementTables():
                if not table.getFileName():
                    table.setFileName(f"chron{chron_counter}measurement{table_counter}.csv")
                for v in table.getVariables():
                    if not v.getVariableId():
                        v.setVariableId(self._generate_unique_id(prefix='TS'))
                table_counter += 1

            model_counter = 0
            for model in chron.getModeledBy():
                table_counter = 0
                for table in model.getEnsembleTables():
                    if not table.getFileName():
                        table.setFileName(f"chron{chron_counter}model{model_counter}ensemble{table_counter}.csv")
                    for v in table.getVariables():
                        if not v.getVariableId():
                            v.setVariableId(self._generate_unique_id(prefix='TS'))
                    table_counter += 1
                model_counter += 1
            chron_counter += 1        


    def load_datasets(self, datasets: 'list[Dataset]'):
        '''
        Loads instances of Dataset class into the LiPD graph

        Parameters
        ----------
        list of pylipd.classes.Dataset
            A list of Dataset objects

        Examples
        --------
        
        pyLipd ships with existing datasets that can be loaded directly through the package. Let's load the Pages2k sample datasets using this method.
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir

            lipd = load_dir('Pages2k')
            dses = lipd.get_datasets()

            # Modify the datasets if needed, then write them to the same, or another LiPD object

            lipd2 = LiPD()
            lipd2.load_datasets(dses)
        
        '''
        for ds in datasets:
            self._fix_missing_ids(ds)
            dsuri = ds.id
            j2r = JSONToRDF(self.graph, dsuri)
            j2r.load_data_in_graph(ds.to_data())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 17:56:50 2023

@author: deborahkhider

Makes it easy to import some specific lpd files for testing and documentation purpose
"""

from pathlib import Path
import glob
from pylipd.lipd import LiPD

DATA_DIR = Path(__file__).parents[1].joinpath("data").resolve()
FOLDER_DIR = DATA_DIR.joinpath('Pages2k/')
TEMP12_DIR = DATA_DIR.joinpath('Temp12k')

def available_dataset_names():
    '''Helper function to easily see what datasets are available to load

    Returns
    -------
    names : list
        List of datasets available via the `load_dataset` method. 

    '''
    path_name = str(DATA_DIR)+'/*.lpd'
    files_unique = glob.glob(path_name)
    
    dir_name = str(FOLDER_DIR)+'/*.lpd'
    files_dir = (glob.glob(dir_name))
    
    temp12_dir = str(TEMP12_DIR)+'/*.lpd'
    temp12_files = (glob.glob(temp12_dir)) 
    
    files = files_unique + files_dir + temp12_files
    
    names = []
    for item in files:
        names.append(item.split('/')[-1].rsplit('.', 1)[0])
    
    return names


def load_datasets(names):
    
    if type(names) is not list:
        names = [names]
    
    path_name = str(DATA_DIR)+'/*.lpd'
    files_unique = glob.glob(path_name)
    dir_name = str(FOLDER_DIR)+'/*.lpd'
    files_dir = (glob.glob(dir_name)) 
    temp12_dir = str(TEMP12_DIR)+'/*.lpd'
    temp12_files = (glob.glob(temp12_dir)) 
    
       
    files = files_unique + files_dir + temp12_files
    
    full_paths = []
    for name in names:
       try:
           full_paths.append(list(filter(lambda a: name in a,files))[0])
       except:
           pass
    
    L = LiPD()
    print(full_paths)
    L.load(full_paths)
   
    return L

def load_dir(name = 'Pages2k'):
    
        
    L = LiPD()
    if name == 'Pages2k':
        L.load_from_dir(str(FOLDER_DIR))
    elif name == 'Temp12k':
        L.load_from_dir(str(TEMP12_DIR))
    else:
        raise ValueError('Directory should be either Pages2k or Temp12k')
    
    return L



"""
The RDF Graph class contains an `RDF <https://www.w3.org/RDF/>`_ Graph using the RDFLib library, and allows querying over it using SPARQL. 
It also allows querying over a remote endpoint.
"""

from copy import deepcopy
import re
from rdflib import ConjunctiveGraph, Namespace, URIRef

from ..globals.urls import ONTONS
from ..utils import sparql_results_to_df

class RDFGraph:
    '''
    The RDF Graph class contains an `RDF <https://www.w3.org/RDF/>`_ Graph using the RDFLib library, and allows querying over it
    
    Examples
    --------
    
    .. jupyter-execute::

        from pylipd.utils.rdf_graph import RDFGraph

        # Load RDF file into graph
        rdf = RDFGraph()
        rdf.load(["../examples/rdf/graph.ttl"])
        (result, result_df) = rdf.query("SELECT ?s ?p ?o WHERE {?s ?p ?o} LIMIT 10")
        print(result_df)    

    '''

    def __init__(self, graph=None):
        if graph is None:
            self._initialize_graph()
        else:
            self.graph = graph

    def _initialize_graph(self):
        self.graph = ConjunctiveGraph()
        self.graph.bind("le", Namespace(ONTONS))        
        #self.graph.bind("", Namespace(NS))
    
    def load(self, files, graphid=None):
        '''Loads a RDF file into the graph

        Parameters
        ----------

        rdf_file : str
            Path to the RDF file

        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Load RDF file into graph
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"])
            (result, result_df) = rdf.query("SELECT ?s ?p ?o WHERE {?s ?p ?o} LIMIT 10")
            print(result_df)
        '''

        for file in files:
            self.graph.parse(file, publicID=graphid)


    def clear(self):
        '''Clears the graph'''
        self._initialize_graph()


    def copy(self):
        '''
        Makes a copy of the object

        Returns
        -------
        pylipd.utils.rdf_graph.RDFGraph
            a copy of the original object

        '''
        
        return deepcopy(self)

    def merge(self, rdf):
        '''
        Merges the current LiPD object with another LiPD object

        Parameters
        ----------
        rdf : pylipd.rdf_graph.RDFGraph
            RDFGraph object to merge with

        Returns
        -------
        pylipd.utils.rdf_graph.RDFGraph
            merged RDFGraph object

        '''

        merged = self.copy()
        merged.graph.addN(rdf.graph.quads())
        return merged
    
    def set_endpoint(self, endpoint):
        '''Sets a SparQL endpoint for a remote Knowledge Base (example: GraphDB)

        Parameters
        ----------

        endpoint : str
            URL for the SparQL endpoint 

        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Fetch LiPD data from remote RDF Graph
            rdf = RDFGraph()
            rdf.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")
            (result, result_df) = rdf.query("SELECT ?s ?p ?o WHERE {?s ?p ?o} LIMIT 10")

        '''
        self.endpoint = endpoint


    def query(self, query, remote=False, result="sparql"):
        '''Once data is loaded into the graph (or remote endpoint set), one can make SparQL queries to the graph

        Parameters
        ----------

        query : str
            SparQL query

        remote: bool
            (Optional) If set to True, the query will be made to the remote endpoint (if set)

        result : str
            (Optional) Result return type

        Returns
        -------

        result : dict
            Dictionary of sparql variable and binding values
        
        result_df : pandas.Dataframe
            Return the dictionary above as a pandas.Dataframe
    
        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"])
            query = """PREFIX le: <http://linked.earth/ontology#>
                    select ?s ?p ?o where { 
                        ?s ?p ?o 
                    } LIMIT 10 """
            result, result_df = rdf.query(query)
            print(result_df)
        '''

        if remote and self.endpoint:
            print("Making remote query to endpoint: " + self.endpoint)
            matches = re.match(r"(.*)\s*SELECT\s+(.+)\s+WHERE\s+{(.+)}\s*(.*)", query, re.DOTALL | re.IGNORECASE)
            if matches:
                prefix = matches.group(1)
                vars = matches.group(2)
                where = matches.group(3)
                suffix = matches.group(4)
                query = f"{prefix} SELECT {vars} WHERE {{ SERVICE <{self.endpoint}> {{ {where} }} }} {suffix}"   
        
        result = self.graph.query(query)
        result_df = sparql_results_to_df(result)
        
        return result, result_df 

    def update(self, update_query, remote=False):
        '''Execute a SPARQL UPDATE query on the graph

        Parameters
        ----------

        update_query : str
            SPARQL UPDATE query (INSERT, DELETE, etc.)

        remote: bool
            (Optional) If set to True, the update will be made to the remote endpoint (if set)

        Returns
        -------

        None

        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"])
            update_query = """PREFIX le: <http://linked.earth/ontology#>
                            DELETE { ?var le:hasMinValue ?oldMin }
                            INSERT { ?var le:hasMinValue "10.5"^^xsd:double }
                            WHERE { ?var le:hasMinValue ?oldMin }"""
            rdf.update(update_query)
        '''

        if remote and hasattr(self, 'endpoint') and self.endpoint:
            print("Making remote update to endpoint: " + self.endpoint)
            # For remote updates, you might need to use a different approach
            # depending on the endpoint's capabilities
            raise NotImplementedError("Remote SPARQL UPDATE not yet implemented")
        
        try:
            self.graph.update(update_query)
        except Exception as e:
            raise Exception(f"SPARQL UPDATE failed: {e}")
    
    def remove(self, ids):
        '''Removes ids(s) from the graph

        Parameters
        ----------

        ids : str or list of str
            graph id(s) to be removed

        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Remove RDF graph data for given id(s)
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"], graphid="http://example.org/graph")
            rdf.remove("http://example.org/graph")
        '''
        
        if type(ids) is not list:
            ids = [ids]

        # Match subgraphs
        for ctx in self.graph.contexts():
            id = str(ctx.identifier)
            if id in ids:
                self.graph.remove((None, None, None, id))       


    def get(self, ids):
        '''Get id(s) from the graph and returns the LiPD object

        Parameters
        ----------

        ids : str or list of str
            graph id(s) to get.

        Returns
        -------

        pylipd.utils.utils.rdf_graph.RDFGraph
            RDFGraph object with the retrieved graph(s)

        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Fetch RDF graph data for given id(s)
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"], graphid="http://example.org/graph")
            rdf.get("http://example.org/graph")  
        '''

        graphds = RDFGraph()

        if type(ids) is not list:
            ids = [ids]

        # Match subgraphs
        for ctx in self.graph.contexts():
            id = str(ctx.identifier)
            if id in ids:
                for triple in self.graph.triples((None, None, None, id)):
                    graphds.graph.add((
                        triple[0],
                        triple[1],
                        triple[2],
                        URIRef(id)))
        return graphds
    

    def pop(self, ids):
        '''Pops graph(s) from the combined graph and returns the popped RDF Graph

        Parameters
        ----------

        ids : str or list of str
            rdf id(s) to be popped.

        Returns
        -------

        pylipd.utils.rdf_graph.RDFGraph
            RDFGraph object with the popped graph(s)

        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Pop RDF graph data for given id(s)
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"], graphid="http://example.org/graph")
            popped = rdf.pop("http://example.org/graph")      
        '''

        popped = self.get(ids)
        self.remove(ids)
        
        return popped
    

    def get_all_graph_ids(self):
        '''Get all Graph ids
        
        Returns
        -------
        
        ids : list
        
        A list of graph ids
        
        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Fetch RDF Graph Data
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"], graphid="http://example.org/graph")
            print(rdf.get_all_graph_ids())
        '''        
        ids = [str(ctx.identifier) for ctx in self.graph.contexts()]
        return ids
    

    def serialize(self):
        '''Returns RDF quad serialization of the current combined Graph
        Examples
        --------

        .. jupyter-execute::

            from pylipd.utils.rdf_graph import RDFGraph

            # Fetch RDF data
            rdf = RDFGraph()
            rdf.load(["../examples/rdf/graph.ttl"], graphid="http://example.org/graph")
            nquads = rdf.serialize()
            print(nquads[:10000])
            print("...")
        '''
        
        return self.graph.serialize(format='nquads')



##############################
# Auto-generated. Do not Edit
##############################
from pylipd.globals.synonyms import SYNONYMS

class PaleoProxy:
    """Controlled-vocabulary class for `PaleoProxy` terms."""
    synonyms = SYNONYMS["PROXIES"]["PaleoProxy"]

    def __init__(self, id, label):
        """Initialize a PaleoProxy term.

        Parameters
        ----------
        id : str
            The full URI identifier for this term.
        label : str
            The human-readable label for this term.
        """
        self.id = id
        self.label = label
    
    def __eq__(self, value: object) -> bool:
        """Check equality based on term ID.

        Parameters
        ----------
        value : object
            The object to compare against.

        Returns
        -------
        bool
            True if the IDs match, False otherwise.
        """
        return self.id == value.id
        
    def getLabel(self):
        """Return the human-readable label of the term.

        Returns
        -------
        str
            The label for this term.
        """
        return self.label

    def getId(self):
        """Return the full URI identifier of the term.

        Returns
        -------
        str
            The URI identifier for this term.
        """
        return self.id
    
    def to_data(self, data={}):
        """Serialize this term to JSON-LD format.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] ={
            "label": [
                {
                    "@datatype": None,
                    "@type": "literal",
                    "@value": self.label
                }
            ]
        }
        return data

    def to_json(self):
        """Return the plain-text label (used in lightweight JSON).

        Returns
        -------
        str
            The label for this term.
        """
        data = self.label
        return data

    @classmethod
    def from_synonym(cls, synonym):
        """Create a PaleoProxy instance from a synonym string.

        Parameters
        ----------
        synonym : str
            A synonym or alternative name for the term.

        Returns
        -------
        PaleoProxy or None
            The PaleoProxy instance if found, None otherwise.
        """
        if synonym.lower() in PaleoProxy.synonyms:
            synobj = PaleoProxy.synonyms[synonym.lower()]
            return PaleoProxy(synobj['id'], synobj['label'])
        return None
        
class PaleoProxyConstants:
    accumulation_rate = PaleoProxy("http://linked.earth/ontology/paleo_proxy#accumulation_rate", "accumulation rate")
    ACL = PaleoProxy("http://linked.earth/ontology/paleo_proxy#ACL", "ACL")
    Al2O3 = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Al2O3", "Al2O3")
    alkenone = PaleoProxy("http://linked.earth/ontology/paleo_proxy#alkenone", "alkenone")
    amoeba = PaleoProxy("http://linked.earth/ontology/paleo_proxy#amoeba", "amoeba")
    Ba_Al = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ba_Al", "Ba/Al")
    Ba_Ca = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ba_Ca", "Ba/Ca")
    biomarker = PaleoProxy("http://linked.earth/ontology/paleo_proxy#biomarker", "biomarker")
    BIT = PaleoProxy("http://linked.earth/ontology/paleo_proxy#BIT", "BIT")
    borehole = PaleoProxy("http://linked.earth/ontology/paleo_proxy#borehole", "borehole")
    BSi = PaleoProxy("http://linked.earth/ontology/paleo_proxy#BSi", "BSi")
    bubble_frequency = PaleoProxy("http://linked.earth/ontology/paleo_proxy#bubble_frequency", "bubble frequency")
    bulk_density = PaleoProxy("http://linked.earth/ontology/paleo_proxy#bulk_density", "bulk density")
    bulk_sediment = PaleoProxy("http://linked.earth/ontology/paleo_proxy#bulk_sediment", "bulk sediment")
    C_N = PaleoProxy("http://linked.earth/ontology/paleo_proxy#C_N", "C/N")
    Ca_K = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ca_K", "Ca/K")
    Ca_Ti = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ca_Ti", "Ca/Ti")
    CaCO3 = PaleoProxy("http://linked.earth/ontology/paleo_proxy#CaCO3", "CaCO3")
    calcification_rate = PaleoProxy("http://linked.earth/ontology/paleo_proxy#calcification_rate", "calcification rate")
    calcite = PaleoProxy("http://linked.earth/ontology/paleo_proxy#calcite", "calcite")
    carbonate = PaleoProxy("http://linked.earth/ontology/paleo_proxy#carbonate", "carbonate")
    cellulose = PaleoProxy("http://linked.earth/ontology/paleo_proxy#cellulose", "cellulose")
    charcoal = PaleoProxy("http://linked.earth/ontology/paleo_proxy#charcoal", "charcoal")
    chironomid = PaleoProxy("http://linked.earth/ontology/paleo_proxy#chironomid", "chironomid")
    chlorophyll = PaleoProxy("http://linked.earth/ontology/paleo_proxy#chlorophyll", "chlorophyll")
    chrysophyte_assemblage = PaleoProxy("http://linked.earth/ontology/paleo_proxy#chrysophyte_assemblage", "chrysophyte assemblage")
    cladoceran = PaleoProxy("http://linked.earth/ontology/paleo_proxy#cladoceran", "cladoceran")
    coccolithophore = PaleoProxy("http://linked.earth/ontology/paleo_proxy#coccolithophore", "coccolithophore")
    d13C = PaleoProxy("http://linked.earth/ontology/paleo_proxy#d13C", "d13C")
    d15N = PaleoProxy("http://linked.earth/ontology/paleo_proxy#d15N", "d15N")
    d15N_d40Ar = PaleoProxy("http://linked.earth/ontology/paleo_proxy#d15N_d40Ar", "d15N/d40Ar")
    d18O = PaleoProxy("http://linked.earth/ontology/paleo_proxy#d18O", "d18O")
    dD = PaleoProxy("http://linked.earth/ontology/paleo_proxy#dD", "dD")
    deuterium_excess = PaleoProxy("http://linked.earth/ontology/paleo_proxy#deuterium_excess", "deuterium excess")
    diatom = PaleoProxy("http://linked.earth/ontology/paleo_proxy#diatom", "diatom")
    dinocyst = PaleoProxy("http://linked.earth/ontology/paleo_proxy#dinocyst", "dinocyst")
    dry_bulk_density = PaleoProxy("http://linked.earth/ontology/paleo_proxy#dry_bulk_density", "dry bulk density")
    Eu_Zr = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Eu_Zr", "Eu/Zr")
    Fe = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Fe", "Fe")
    Fe_Al = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Fe_Al", "Fe/Al")
    foraminifera = PaleoProxy("http://linked.earth/ontology/paleo_proxy#foraminifera", "foraminifera")
    GDGT = PaleoProxy("http://linked.earth/ontology/paleo_proxy#GDGT", "GDGT")
    grain_size = PaleoProxy("http://linked.earth/ontology/paleo_proxy#grain_size", "grain size")
    HBI = PaleoProxy("http://linked.earth/ontology/paleo_proxy#HBI", "HBI")
    historical = PaleoProxy("http://linked.earth/ontology/paleo_proxy#historical", "historical")
    humification = PaleoProxy("http://linked.earth/ontology/paleo_proxy#humification", "humification")
    ice_accumulation = PaleoProxy("http://linked.earth/ontology/paleo_proxy#ice_accumulation", "ice accumulation")
    ice_melt = PaleoProxy("http://linked.earth/ontology/paleo_proxy#ice_melt", "ice melt")
    inorganic_carbon = PaleoProxy("http://linked.earth/ontology/paleo_proxy#inorganic_carbon", "inorganic carbon")
    IP25 = PaleoProxy("http://linked.earth/ontology/paleo_proxy#IP25", "IP25")
    lake_level = PaleoProxy("http://linked.earth/ontology/paleo_proxy#lake_level", "lake level")
    latewood_cellulose = PaleoProxy("http://linked.earth/ontology/paleo_proxy#latewood_cellulose", "latewood cellulose")
    LDI = PaleoProxy("http://linked.earth/ontology/paleo_proxy#LDI", "LDI")
    macrofossils = PaleoProxy("http://linked.earth/ontology/paleo_proxy#macrofossils", "macrofossils")
    magnetic = PaleoProxy("http://linked.earth/ontology/paleo_proxy#magnetic", "magnetic")
    magnetic_susceptibility = PaleoProxy("http://linked.earth/ontology/paleo_proxy#magnetic_susceptibility", "magnetic susceptibility")
    mass_accumulation_rate = PaleoProxy("http://linked.earth/ontology/paleo_proxy#mass_accumulation_rate", "mass accumulation rate")
    maximum_latewood_density = PaleoProxy("http://linked.earth/ontology/paleo_proxy#maximum_latewood_density", "maximum latewood density")
    Mg = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Mg", "Mg")
    Mg_Ca = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Mg_Ca", "Mg/Ca")
    multiproxy = PaleoProxy("http://linked.earth/ontology/paleo_proxy#multiproxy", "multiproxy")
    Ti = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ti", "Ti")
    needs_to_be_changed = PaleoProxy("http://linked.earth/ontology/paleo_proxy#needs_to_be_changed", "needs to be changed")
    needsToBeChanged = PaleoProxy("http://linked.earth/ontology/paleo_proxy#needsToBeChanged", "needsToBeChanged")
    ostracod = PaleoProxy("http://linked.earth/ontology/paleo_proxy#ostracod", "ostracod")
    P_aqueous = PaleoProxy("http://linked.earth/ontology/paleo_proxy#P-aqueous", "P-aqueous")
    peat_ash = PaleoProxy("http://linked.earth/ontology/paleo_proxy#peat_ash", "peat ash")
    pH = PaleoProxy("http://linked.earth/ontology/paleo_proxy#pH", "pH")
    pollen = PaleoProxy("http://linked.earth/ontology/paleo_proxy#pollen", "pollen")
    radiolaria = PaleoProxy("http://linked.earth/ontology/paleo_proxy#radiolaria", "radiolaria")
    Rb = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Rb", "Rb")
    Rb_Sr = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Rb_Sr", "Rb/Sr")
    reflectance = PaleoProxy("http://linked.earth/ontology/paleo_proxy#reflectance", "reflectance")
    ring_width = PaleoProxy("http://linked.earth/ontology/paleo_proxy#ring_width", "ring width")
    Sr = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Sr", "Sr")
    Sr_Ca = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Sr_Ca", "Sr/Ca")
    stratigraphy = PaleoProxy("http://linked.earth/ontology/paleo_proxy#stratigraphy", "stratigraphy")
    sulfur = PaleoProxy("http://linked.earth/ontology/paleo_proxy#sulfur", "sulfur")
    TEX86 = PaleoProxy("http://linked.earth/ontology/paleo_proxy#TEX86", "TEX86")
    Ti_Al = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ti_Al", "Ti/Al")
    Ti_Ca = PaleoProxy("http://linked.earth/ontology/paleo_proxy#Ti_Ca", "Ti/Ca")
    TOC = PaleoProxy("http://linked.earth/ontology/paleo_proxy#TOC", "TOC")
    total_nitrogen = PaleoProxy("http://linked.earth/ontology/paleo_proxy#total_nitrogen", "total nitrogen")
    varve_thickness = PaleoProxy("http://linked.earth/ontology/paleo_proxy#varve_thickness", "varve thickness")



##############################
# Auto-generated. Do not Edit
##############################
from pylipd.globals.synonyms import SYNONYMS

class InterpretationSeasonality:
    """Controlled-vocabulary class for `InterpretationSeasonality` terms."""
    synonyms = SYNONYMS["INTERPRETATION"]["InterpretationSeasonality"]

    def __init__(self, id, label):
        """Initialize a InterpretationSeasonality term.

        Parameters
        ----------
        id : str
            The full URI identifier for this term.
        label : str
            The human-readable label for this term.
        """
        self.id = id
        self.label = label
    
    def __eq__(self, value: object) -> bool:
        """Check equality based on term ID.

        Parameters
        ----------
        value : object
            The object to compare against.

        Returns
        -------
        bool
            True if the IDs match, False otherwise.
        """
        return self.id == value.id
        
    def getLabel(self):
        """Return the human-readable label of the term.

        Returns
        -------
        str
            The label for this term.
        """
        return self.label

    def getId(self):
        """Return the full URI identifier of the term.

        Returns
        -------
        str
            The URI identifier for this term.
        """
        return self.id
    
    def to_data(self, data={}):
        """Serialize this term to JSON-LD format.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] ={
            "label": [
                {
                    "@datatype": None,
                    "@type": "literal",
                    "@value": self.label
                }
            ]
        }
        return data

    def to_json(self):
        """Return the plain-text label (used in lightweight JSON).

        Returns
        -------
        str
            The label for this term.
        """
        data = self.label
        return data

    @classmethod
    def from_synonym(cls, synonym):
        """Create a InterpretationSeasonality instance from a synonym string.

        Parameters
        ----------
        synonym : str
            A synonym or alternative name for the term.

        Returns
        -------
        InterpretationSeasonality or None
            The InterpretationSeasonality instance if found, None otherwise.
        """
        if synonym.lower() in InterpretationSeasonality.synonyms:
            synobj = InterpretationSeasonality.synonyms[synonym.lower()]
            return InterpretationSeasonality(synobj['id'], synobj['label'])
        return None
        
class InterpretationSeasonalityConstants:
    Annual = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Annual", "Annual")
    Winter = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Winter", "Winter")
    Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr", "Apr")
    Oct_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-May", "Oct-May")
    Apr_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Aug", "Apr-Aug")
    Apr_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Dec", "Apr-Dec")
    Apr_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Feb", "Apr-Feb")
    Apr_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Jan", "Apr-Jan")
    Apr_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Jul", "Apr-Jul")
    Apr_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Jun", "Apr-Jun")
    Apr_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Mar", "Apr-Mar")
    Apr_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-May", "Apr-May")
    Apr_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Nov", "Apr-Nov")
    Apr_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Oct", "Apr-Oct")
    Apr_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Apr-Sep", "Apr-Sep")
    Summer = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Summer", "Summer")
    Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug", "Aug")
    Aug_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Apr", "Aug-Apr")
    Aug_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Dec", "Aug-Dec")
    Aug_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Feb", "Aug-Feb")
    Aug_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Jan", "Aug-Jan")
    Aug_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Jul", "Aug-Jul")
    Aug_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Jun", "Aug-Jun")
    Aug_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Mar", "Aug-Mar")
    Aug_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-May", "Aug-May")
    Aug_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Nov", "Aug-Nov")
    Aug_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Oct", "Aug-Oct")
    Aug_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Aug-Sep", "Aug-Sep")
    Growing_Season = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Growing_Season", "Growing Season")
    Coldest_Month = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Coldest_Month", "Coldest Month")
    Dec_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Apr", "Dec-Apr")
    Dec_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Aug", "Dec-Aug")
    Dec_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Feb", "Dec-Feb")
    Dec_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Jan", "Dec-Jan")
    Dec_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Jul", "Dec-Jul")
    Dec_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Jun", "Dec-Jun")
    Dec_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Mar", "Dec-Mar")
    Dec_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-May", "Dec-May")
    Dec_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Oct", "Dec-Oct")
    Dec_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Dec-Sep", "Dec-Sep")
    Fall = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Fall", "Fall")
    Feb_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Aug", "Feb-Aug")
    Feb_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Apr", "Feb-Apr")
    Feb_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Dec", "Feb-Dec")
    Feb_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Jul", "Feb-Jul")
    Feb_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Jun", "Feb-Jun")
    Feb_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Mar", "Feb-Mar")
    Feb_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-May", "Feb-May")
    Feb_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Nov", "Feb-Nov")
    Feb_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Oct", "Feb-Oct")
    Feb_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Feb-Sep", "Feb-Sep")
    Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan", "Jan")
    Jan_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Apr", "Jan-Apr")
    Jan_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Aug", "Jan-Aug")
    Jan_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Feb", "Jan-Feb")
    Jan_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Jul", "Jan-Jul")
    Jan_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Jun", "Jan-Jun")
    Jan_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Mar", "Jan-Mar")
    Jan_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-May", "Jan-May")
    Jan_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Nov", "Jan-Nov")
    Jan_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Oct", "Jan-Oct")
    Jan_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jan-Sep", "Jan-Sep")
    Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul", "Jul")
    May_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Sep", "May-Sep")
    Jul_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Apr", "Jul-Apr")
    Jul_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Aug", "Jul-Aug")
    Jul_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Dec", "Jul-Dec")
    Jul_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Feb", "Jul-Feb")
    Jul_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Jan", "Jul-Jan")
    Jul_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Jun", "Jul-Jun")
    Jul_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Mar", "Jul-Mar")
    Jul_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-May", "Jul-May")
    Jul_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Nov", "Jul-Nov")
    Jul_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Oct", "Jul-Oct")
    Jul_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jul-Sep", "Jul-Sep")
    Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun", "Jun")
    Jun_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Apr", "Jun-Apr")
    Jun_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Aug", "Jun-Aug")
    Jun_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Sep", "Jun-Sep")
    Jun_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Dec", "Jun-Dec")
    Jun_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Feb", "Jun-Feb")
    Jun_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Jan", "Jun-Jan")
    Jun_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Jul", "Jun-Jul")
    Jun_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Mar", "Jun-Mar")
    Jun_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Nov", "Jun-Nov")
    Jun_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Jun-Oct", "Jun-Oct")
    Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar", "Mar")
    Mar_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Apr", "Mar-Apr")
    Mar_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Aug", "Mar-Aug")
    Mar_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Dec", "Mar-Dec")
    Mar_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Jan", "Mar-Jan")
    Mar_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Jul", "Mar-Jul")
    Mar_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Jun", "Mar-Jun")
    Mar_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-May", "Mar-May")
    Mar_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Nov", "Mar-Nov")
    Mar_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Oct", "Mar-Oct")
    Mar_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Mar-Sep", "Mar-Sep")
    May_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Apr", "May-Apr")
    May_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Aug", "May-Aug")
    May_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Dec", "May-Dec")
    May_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Oct", "May-Oct")
    May_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Feb", "May-Feb")
    May_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Jan", "May-Jan")
    May_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Jul", "May-Jul")
    May_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Jun", "May-Jun")
    May_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Mar", "May-Mar")
    May_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#May-Nov", "May-Nov")
    needsToBeChanged = InterpretationSeasonality("http://linked.earth/ontology/interpretation#needsToBeChanged", "needsToBeChanged")
    Nov_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Apr", "Nov-Apr")
    Nov_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Aug", "Nov-Aug")
    Nov_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Dec", "Nov-Dec")
    Nov_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Feb", "Nov-Feb")
    Nov_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Jan", "Nov-Jan")
    Nov_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Jul", "Nov-Jul")
    Nov_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Jun", "Nov-Jun")
    Nov_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Mar", "Nov-Mar")
    Nov_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-May", "Nov-May")
    Nov_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Oct", "Nov-Oct")
    Nov_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Nov-Sep", "Nov-Sep")
    Oct_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Apr", "Oct-Apr")
    Oct_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Aug", "Oct-Aug")
    Oct_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Dec", "Oct-Dec")
    Oct_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Feb", "Oct-Feb")
    Oct_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Jan", "Oct-Jan")
    Oct_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Jul", "Oct-Jul")
    Oct_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Jun", "Oct-Jun")
    Oct_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Mar", "Oct-Mar")
    Oct_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Nov", "Oct-Nov")
    Oct_Sep = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Oct-Sep", "Oct-Sep")
    Sep_Apr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Apr", "Sep-Apr")
    Sep_Aug = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Aug", "Sep-Aug")
    Sep_Dec = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Dec", "Sep-Dec")
    Sep_Feb = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Feb", "Sep-Feb")
    Sep_Jan = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Jan", "Sep-Jan")
    Sep_Jul = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Jul", "Sep-Jul")
    Sep_Jun = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Jun", "Sep-Jun")
    Sep_Mar = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Mar", "Sep-Mar")
    Sep_May = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-May", "Sep-May")
    Sep_Nov = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Nov", "Sep-Nov")
    Sep_Oct = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Sep-Oct", "Sep-Oct")
    Spr_Sum = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Spr-Sum", "Spr-Sum")
    Spring = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Spring", "Spring")
    subannual = InterpretationSeasonality("http://linked.earth/ontology/interpretation#subannual", "subannual")
    Warmest_Month = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Warmest_Month", "Warmest Month")
    Wet_Season = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Wet_Season", "Wet Season")
    Win_Spr = InterpretationSeasonality("http://linked.earth/ontology/interpretation#Win-Spr", "Win-Spr")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

@author: deborahkhider

Contains all basic block functions for pylipd. 

"""

from .utils import *
from .legacy_utils import *
from .lipd_to_rdf import *
from .multi_processing import *
from .rdf_graph import *
from .rdf_to_lipd import *
from .dataset import *


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 19 12:18:41 2025

@author: deborahkhider

Tests for the LiPD classes. 

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}
Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
"""

import pytest
from pylipd.classes.dataset import Dataset
from pylipd.lipd import LiPD
import pandas as pd
import numpy as np

from pylipd.classes.archivetype import ArchiveTypeConstants
from pylipd.classes.funding import Funding
from pylipd.classes.interpretation import Interpretation
from pylipd.classes.interpretationvariable import InterpretationVariableConstants
from pylipd.classes.location import Location
from pylipd.classes.paleodata import PaleoData
from pylipd.classes.datatable import DataTable
from pylipd.classes.paleounit import PaleoUnitConstants
from pylipd.classes.paleovariable import PaleoVariableConstants
from pylipd.classes.person import Person
from pylipd.classes.publication import Publication
from pylipd.classes.resolution import Resolution
from pylipd.classes.variable import Variable

import json

import re

import uuid

def generate_unique_id(prefix='PYD'):
    # Generate a random UUID
    random_uuid = uuid.uuid4()  # Generates a random UUID.
    
    # Convert UUID format to the specific format we need
    # UUID is usually in the form '1e2a2846-2048-480b-9ec6-674daef472bd' so we slice and insert accordingly
    id_str = str(random_uuid)
    formatted_id = f"{prefix}-{id_str[:5]}-{id_str[9:13]}-{id_str[14:18]}-{id_str[19:23]}-{id_str[24:28]}"
    
    return formatted_id


class TestGet():
    """
    This class test the get information function. Let's use the ODP846 dataset as it contains ensemble tables
    """

    def test_get_dataset(self, odp846):
        D = odp846
        D.get_datasets()
    
    def test_get_name(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        name = ds.getName()
        assert name == 'ODP846.Lawrence.2006'
    
    def test_get_funding(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        funding = ds.getFundings()
        assert len(funding) == 0
    
    def test_get_publication(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        pub = ds.getPublications()
        assert len(pub) == 6
    
    def test_get_pub_info(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        pub = ds.getPublications()
        pub1 = pub[0]
        
        pub1.getTitle()
        authors = pub1.getAuthors()
        authors[0].getName()
        pub1.getJournal()
        pub1.getDOI()
        pub1.getPages()
        pub1.getIssue()
        pub1.getVolume()
        pub1.getYear()
    
    def test_get_location(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        geo = ds.getLocation()
        lat = geo.getLatitude()
        lon = geo.getLongitude()
        elev = geo.getElevation()
        coor = geo.getCoordinates()
        assert lat == -3.1
        assert lon == -90.8
        assert elev == -3296.0
        assert coor == '-3.1,-90.8'
    
    def test_get_paleodata(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        paleodata = ds.getPaleoData()
        assert len(paleodata) == 1
    
    def test_get_chrondata(self,odp846):
        D = odp846
        ds = D.get_datasets()[0]
        chrondata = ds.getChronData()
        assert len(chrondata) == 1
    
    def test_get_datatable(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        paleodata = ds.getPaleoData()
        data_tables = []
        for item in paleodata:
            for table in item.getMeasurementTables(): #get the measurement tables
                df = table.getDataFrame(use_standard_names=True) # grab the data and standardize the variable names
                data_tables.append(df)
        assert len(data_tables) == 2
    
    def test_get_model(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        chrondata = ds.getChronData()
        df_ens = [] # create an empty list to store all ensemble tables across models. 
        for item in chrondata:
            for model in item.getModeledBy():
                for etable in model.getEnsembleTables():
                    df_ens.append(etable.getDataFrame())
        assert len(df_ens) == 1
    
    def test_get_table(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        paleodata = ds.getPaleoData()
        table = paleodata[0].getMeasurementTables()[0]
        
        table.getFileName()
        table.getMissingValue()
              
        var = table.getVariables()
        assert len(var) == 9
        
    def test_get_variable(self, odp846):
        D = odp846
        ds = D.get_datasets()[0]
        paleodata = ds.getPaleoData()
        table = paleodata[0].getMeasurementTables()[0]
        var = table.getVariables()
        var1 = var[0]
        
        var1.getColumnNumber()
        var1.getDescription()
        var1.getFoundInDataset()
        var1.getFoundInTable()
        var1.getInterpretations()
        var1.getMaxValue()
        var1.getMeanValue()
        var1.getMedianValue()
        var1.getMinValue()
        var1.getMissingValue()
        var1.getName()
        var1.getNotes()
        var1.getPartOfCompilations()
        var1.getProxy()
        var1.getProxyGeneral()
        var1.getResolution()
        var1.getStandardVariable()
        var1.getUnits()
        var1.getValues()
        var1.getVariableId()

class TestCreate():
    """
    Test the functionalities to create/edit a LiPD file
    """
    
    def test_create_ds(self):
        
        dataset1 = Dataset()

        # Set the name of the dataset
        dataset1.setName("TestDataset.2024")
        dataset1.id = dataset1.ns + "/" + dataset1.getName() # ** IMPORTANT **
        
        # Set collection name
        dataset1.setCollectionName("TestCollection")
        
        # Set the Archive Type (from a list of constants)
        dataset1.setArchiveType(ArchiveTypeConstants.MarineSediment)
        
        # Add a publication
        pub1 = Publication()
        pub1.setTitle("Sample Publication Title")
        person1 = Person(); person1.setName("Deborah Khider")
        person2 = Person(); person2.setName("Varun Ratnakar")
        pub1.setAuthors([person1, person2])
        # Add the publication to the dataset
        dataset1.addPublication(pub1)
        
        # Add funding information
        funding1 = Funding()
        funding1.addGrant("NSF Grant 23423A")
        funding1.addInvestigator(person1)
        # Add funding to the dataset
        dataset1.addFunding(funding1)
        
        # Add location information
        loc1 = Location()
        loc1.setLatitude("24.21232")
        loc1.setLongitude("48.32323")
        loc1.setElevation("342")
        loc1.setCountry("USA")
        # Set location for the dataset
        dataset1.setLocation(loc1)
        
        # Create Paleodata table
        table1 = DataTable()
        table1.setFileName("paleo0measurement1.csv")
        table1.setMissingValue("NaN")
        
        # Populate the table with variable data
        #
        # Option 1: Via a Dataframe with attributes:
        # ------------------------------------------
        # Create a random dataframe
        df = pd.DataFrame(np.random.randn(100, 2), columns=["site", "ukprime37"])
        # Set column attributes
        df.attrs = {
            "site": {
                'number': 1, 
                'variableName': 'site/hole', 
                'units': 'unitless', 
                'TSid': 'PYTJ3PSH0LT', 
                'variableType': 'measured', 
                'takenAtDepth': 'depth'
            },
            "ukprime37": {
                'number': 2,
                'interpretation': [
                    {
                        'rank': 1.0, 
                        'scope': 'Climate', 
                        'variable': 'temperature', 
                        'variableDetail': 'sea surface', 
                        'direction': 'positive'
                    }
                ], 
                'variableName': 'ukprime37', 
                'resolution': {
                    'hasMaxValue': 10.856999999999971, 
                    'hasMeanValue': 2.3355875057418465, 
                    'hasMedianValue': 2.211999999999989, 
                    'hasMinValue': 0.06999999999993634
                }, 
                'TSid': 'PYTM9N6HCQM', 
                'variableType': 'measured', 
                'takenAtDepth': 'depth' 
            }
        }
        table1.setDataFrame(df)
        
        # Create Another Paleodata table by setting variables with OOP calls
        table2 = DataTable()
        table2.setFileName("paleo0measurement2.csv")
        table2.setMissingValue("NaN")
        #
        # Option 2: Via OOP Calls to create each variable
        # -----------------------------------------------
        # Add a variable
        var1 = Variable()
        var1.setName("site")
        var1.setColumnNumber(1)
        var1.setVariableId("PYTJ3PSH0LT")
        var1.setVariableType("measured")
        var1.set_non_standard_property("takenAtDepth", "depth")
        # Set random values for this variable
        var1.setValues(json.dumps(np.random.randn(100).tolist()))
        
        # Add another variable
        var2 = Variable()
        var2.setName("ukprime37")
        var2.setColumnNumber(2)
        var2.setVariableId('PYTM9N6HCQM')
        var2.setVariableType('measured')
        var2.setStandardVariable(PaleoVariableConstants.Uk37)
        var2.setUnits(PaleoUnitConstants.cm3)
        
        # Add the variable interpretation
        interp1 = Interpretation()
        interp1.setRank("1")
        interp1.setScope("Climate")
        interp1.setVariable(InterpretationVariableConstants.temperature)
        interp1.setVariableDetail("sea surface")
        interp1.setDirection("positive")
        var2.addInterpretation(interp1)
        # Add the variable resolution
        resolution1 = Resolution()
        resolution1.setMaxValue(10.856999999999971)
        resolution1.setMeanValue(2.3355875057418465)
        resolution1.setMedianValue(2.211999999999989)
        resolution1.setMinValue(0.06999999999993634)
        var2.setResolution(resolution1)
        
        # Set random values for this variable
        var2.setValues(json.dumps(np.random.randn(100).tolist()))
        
        table2.setVariables([var1, var2])
        
        # Create Paleodata, and add the created tables to it
        paleodata1 = PaleoData()
        paleodata1.setMeasurementTables([table1, table2])
        
        dataset1.addPaleoData(paleodata1)
        
        
        



##############################
# Auto-generated. Do not Edit
##############################
from pylipd.globals.synonyms import SYNONYMS

class InterpretationVariable:
    """Controlled-vocabulary class for `InterpretationVariable` terms."""
    synonyms = SYNONYMS["INTERPRETATION"]["InterpretationVariable"]

    def __init__(self, id, label):
        """Initialize a InterpretationVariable term.

        Parameters
        ----------
        id : str
            The full URI identifier for this term.
        label : str
            The human-readable label for this term.
        """
        self.id = id
        self.label = label
    
    def __eq__(self, value: object) -> bool:
        """Check equality based on term ID.

        Parameters
        ----------
        value : object
            The object to compare against.

        Returns
        -------
        bool
            True if the IDs match, False otherwise.
        """
        return self.id == value.id
        
    def getLabel(self):
        """Return the human-readable label of the term.

        Returns
        -------
        str
            The label for this term.
        """
        return self.label

    def getId(self):
        """Return the full URI identifier of the term.

        Returns
        -------
        str
            The URI identifier for this term.
        """
        return self.id
    
    def to_data(self, data={}):
        """Serialize this term to JSON-LD format.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] ={
            "label": [
                {
                    "@datatype": None,
                    "@type": "literal",
                    "@value": self.label
                }
            ]
        }
        return data

    def to_json(self):
        """Return the plain-text label (used in lightweight JSON).

        Returns
        -------
        str
            The label for this term.
        """
        data = self.label
        return data

    @classmethod
    def from_synonym(cls, synonym):
        """Create a InterpretationVariable instance from a synonym string.

        Parameters
        ----------
        synonym : str
            A synonym or alternative name for the term.

        Returns
        -------
        InterpretationVariable or None
            The InterpretationVariable instance if found, None otherwise.
        """
        if synonym.lower() in InterpretationVariable.synonyms:
            synobj = InterpretationVariable.synonyms[synonym.lower()]
            return InterpretationVariable(synobj['id'], synobj['label'])
        return None
        
class InterpretationVariableConstants:
    C3C4Ratio = InterpretationVariable("http://linked.earth/ontology/interpretation#C3C4Ratio", "C3C4Ratio")
    circulationIndex = InterpretationVariable("http://linked.earth/ontology/interpretation#circulationIndex", "circulationIndex")
    circulationVariable = InterpretationVariable("http://linked.earth/ontology/interpretation#circulationVariable", "circulationVariable")
    dissolvedOxygen = InterpretationVariable("http://linked.earth/ontology/interpretation#dissolvedOxygen", "dissolvedOxygen")
    dust = InterpretationVariable("http://linked.earth/ontology/interpretation#dust", "dust")
    ELA = InterpretationVariable("http://linked.earth/ontology/interpretation#ELA", "ELA")
    evaporation = InterpretationVariable("http://linked.earth/ontology/interpretation#evaporation", "evaporation")
    fire = InterpretationVariable("http://linked.earth/ontology/interpretation#fire", "fire")
    growingDegreeDays = InterpretationVariable("http://linked.earth/ontology/interpretation#growingDegreeDays", "growingDegreeDays")
    hydrologicBalance = InterpretationVariable("http://linked.earth/ontology/interpretation#hydrologicBalance", "hydrologicBalance")
    lakeWaterIsotope = InterpretationVariable("http://linked.earth/ontology/interpretation#lakeWaterIsotope", "lakeWaterIsotope")
    meltwater = InterpretationVariable("http://linked.earth/ontology/interpretation#meltwater", "meltwater")
    needsToBeReplaced = InterpretationVariable("http://linked.earth/ontology/interpretation#needsToBeReplaced", "needsToBeReplaced")
    P_E = InterpretationVariable("http://linked.earth/ontology/interpretation#P-E", "P-E")
    precipitation = InterpretationVariable("http://linked.earth/ontology/interpretation#precipitation", "precipitation")
    precipitationDeuteriumExcess = InterpretationVariable("http://linked.earth/ontology/interpretation#precipitationDeuteriumExcess", "precipitationDeuteriumExcess")
    precipitationIsotope = InterpretationVariable("http://linked.earth/ontology/interpretation#precipitationIsotope", "precipitationIsotope")
    productivity = InterpretationVariable("http://linked.earth/ontology/interpretation#productivity", "productivity")
    relativeHumidity = InterpretationVariable("http://linked.earth/ontology/interpretation#relativeHumidity", "relativeHumidity")
    salinity = InterpretationVariable("http://linked.earth/ontology/interpretation#salinity", "salinity")
    seaIce = InterpretationVariable("http://linked.earth/ontology/interpretation#seaIce", "seaIce")
    seasonality = InterpretationVariable("http://linked.earth/ontology/interpretation#seasonality", "seasonality")
    seawaterIsotope = InterpretationVariable("http://linked.earth/ontology/interpretation#seawaterIsotope", "seawaterIsotope")
    streamflow = InterpretationVariable("http://linked.earth/ontology/interpretation#streamflow", "streamflow")
    sunlight = InterpretationVariable("http://linked.earth/ontology/interpretation#sunlight", "sunlight")
    surfacePressure = InterpretationVariable("http://linked.earth/ontology/interpretation#surfacePressure", "surfacePressure")
    temperature = InterpretationVariable("http://linked.earth/ontology/interpretation#temperature", "temperature")
    upwelling = InterpretationVariable("http://linked.earth/ontology/interpretation#upwelling", "upwelling")
    windSpeed = InterpretationVariable("http://linked.earth/ontology/interpretation#windSpeed", "windSpeed")